from .platform import Platform
from .organization import Organization
from .organization_list import OrganizationList
from .organization_filter import OrganizationFilter

__all__ = [
    'Platform',
    'Organization',
    'OrganizationList',
    'OrganizationFilter',
]


"""
   Descp: This class is used to represent a calendar plot.

   Created on: 21-apr-2022

   Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""
from typing import Dict

import plotly.graph_objs as go
from plotly_calplot import month_calplot

from .figure import Figure
import dao_analyzer.apps.common.resources.colors as Color

class CalFigure(Figure):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def get_empty_plot_data() -> Dict:
        return {
            'x': [],
            'y': [],
        }
    
    def get_empty_figure(self) -> Dict:
        return self.get_figure(self.get_empty_plot_data())
    
    def get_figure(self, plot_data: Dict) -> Dict:
        fig: go.Figure = month_calplot(
            x=plot_data['x'],
            y=plot_data['y'],
            width=None,
            gap=3,
            year_height=50,
        )

        fig.update_layout(
            margin=self.configuration.get_margin(),
            xaxis = {
                'tickfont_color': Color.TICKFONT_COLOR,
            },
            yaxis = {
                'tickfont_color': Color.TICKFONT_COLOR,
            }
        )

        return fig


"""
   Descp: Platform transfers.

   Created on: 1-jun-2022

   Copyright 2022 David Davó Laviña
        <ddavo@ucm.es>
"""

from typing import List, Dict, Any

from datetime import datetime

from .participation_stats import ParticipationStat

class Platform:
    def __init__(self,
        name: str,
        networks: List[str] = [],
        participation_stats: List[ParticipationStat] = [],
        creation_date: datetime = None,
    ):
        self._name = name
        self._networks = networks
        self._stats = participation_stats
        self._creation_date = creation_date

    @property
    def name(self) -> str:
        return self._name

    @property
    def networks(self) -> List[str]:
        return self._networks

    @property
    def participation_stats(self) -> List[ParticipationStat]:
        return self._stats

    @property
    def creation_date(self) -> datetime:
        return self._creation_date

    def get_dropdown_representation(self) -> List[Dict[str, str]]:
        return self._orgs.get_dict_representation()

    def to_plotly_json(self) -> Dict[str, Any]:
        return {
            'name': str(self.name),
            'networks': list(self.networks),
            'stats': list(self.participation_stats),
            'creation_date': self.creation_date.isoformat() if self.creation_date else None,
        }

    @classmethod
    def from_json(cls, dict: Dict[str, Any]) -> 'Platform':
        if not dict:
            return None
        
        return Platform(
            name = dict['name'],
            networks = dict['networks'],
            participation_stats = list(map(ParticipationStat.from_json, dict.get('stats', []))),
            creation_date = None if dict['creation_date'] == 'NaT' else datetime.fromisoformat(dict['creation_date']),
        )


"""
   Descp: This is a dao (data access object) of metrics.
    It's used as interface to transform data from the datawarehouse
    and transfers like serie, stackedSerie, NStackedSerie.  

   Created on: 17-jul-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List
import pandas as pd

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.apps.common.data_access.requesters.irequester import IRequester


class MetricDao():
    def __init__(self, ids: List[str], strategy: IMetricStrategy,
     requester: IRequester, address_key: str):
        self.__ids = ids
        self.__st = strategy
        self.__requester = requester
        self.__address_key = address_key


    def get_metric(self):
        df: pd.DataFrame = self.__requester.request()
        assert(not df.empty)
        
        # get only data from daos in ids
        df = df.loc[df[self.__address_key].isin(self.__ids)]
        
        # So process_data doesn't modify it
        df = df.copy()

        return self.__st.process_data(df)



"""
   Descp: Strategy pattern to create total reputation holders.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.serie import Serie 
import dao_analyzer.apps.common.data_access.pandas_utils as pd_utl
import dao_analyzer.apps.daostack.data_access.daos.metric.strategy.st_time_serie \
    as new_reputation_holders


class StTotalRepHolders(IMetricStrategy):

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        new_holders: StackedSerie = new_reputation_holders\
            .StTimeSerie(m_type=new_reputation_holders.NEW_USERS).process_data(df)

        total: List[int] = self.__calculate_total(new_holders=new_holders)

        metric: StackedSerie = StackedSerie(
            serie = Serie(x=new_holders.get_serie()), 
            y_stack = [total])

        return metric


    def __calculate_total(self, new_holders: StackedSerie) -> List[int]:
        news: List[int] = new_holders.get_i_stack(0)
        total: List[int] = [news[0]]

        for i in range(1, len(news)):
            total.append(news[i] + total[i-1])

        return total



"""
   Descp: This is a dao (data access object) of the organization.
    It's used to drive the data trasformation from datawarehouse to transfer obj.

   Created on: 22-may-2022

   Copyright 2022-2022 David Davó Laviña
        <david@ddavo.me>
"""
import pandas as pd

from dao_analyzer.apps.common.business.transfers import Organization, OrganizationList
from dao_analyzer.apps.common.business.transfers.organization.participation_stats import MembersCreatedProposalsStat, MembersEverVotedStat
from dao_analyzer.apps.common.business.transfers.organization.platform import Platform
from dao_analyzer.apps.common.data_access.daos.platform_dao import PlatformDAO
from dao_analyzer.apps.common.data_access.requesters import CacheRequester

from .metric import srcs

class DaohausDao(PlatformDAO):
    __DF_IDX = ['network', 'molochAddress']
    __DF_DATE = 'createdAt'
    __DF_PROPOSER = 'memberAddress'
    __DF_MEMBER = 'memberAddress'
    __DF_VOTER = 'memberAddress'
    __DF_ACT_COLS = __DF_IDX + [__DF_DATE]

    __EMPTY_ADDR = '0x0000000000000000000000000000000000000000'

    def __init__(self):
        super().__init__(CacheRequester(srcs=[
            srcs.MOLOCHES,
        ]))

        self._activityRequester = CacheRequester([
            srcs.MOLOCHES,
            srcs.PROPOSALS,
            srcs.RAGE_QUITS,
            srcs.VOTES,
        ])

        self._membersRequester = CacheRequester([
            srcs.MEMBERS,
        ])

        self._votesRequester = CacheRequester([
            srcs.VOTES,
        ])

        self._propRequester = CacheRequester([
            srcs.PROPOSALS,
        ])

    def _get_daos(self) -> pd.DataFrame:
        df = self._requester.request().set_index(self.__DF_IDX, drop=False)

        df['summoningTime'] = pd.to_datetime(df['summoningTime'], unit='s')

        return df

    @staticmethod
    def _filter_by_daos(df, daos) -> pd.DataFrame:
        if daos is None:
            return df
        else:
            return df[df.set_index(DaohausDao.__DF_IDX).index.isin(daos.index)]

    def _get_participation(self, daos: pd.DataFrame = None) -> pd.DataFrame:
        participation = self._propRequester.request()
        participation = participation.replace(self.__EMPTY_ADDR, None)

        participation = self._filter_by_daos(participation, daos)

        return participation

    def _get_activity(self) -> pd.DataFrame:
        activity = self._activityRequester.request()
        activity = activity[self.__DF_ACT_COLS]
        activity[self.__DF_DATE] = pd.to_datetime(activity[self.__DF_DATE], unit='s')

        return activity

    def _get_members(self) -> pd.DataFrame:
        members = self._membersRequester.request()
        members = members.replace(self.__EMPTY_ADDR, None)

        return members

    def _get_votes(self, daos: pd.DataFrame = None) -> pd.DataFrame:
        votes = self._votesRequester.request()

        votes = self._filter_by_daos(votes, daos)

        return votes

    def get_platform(self, orglist: OrganizationList) -> Platform:
        df = self._get_daos()
        members = self._get_members()

        if orglist:
            ids = { (o.get_network(), o.get_id()) for o in orglist }
            
            df = df[df.index.isin(ids)]
            members = self._filter_by_daos(members, df)

        mcp_pct = self._get_participation(daos=df)[self.__DF_PROPOSER].nunique() / members[self.__DF_MEMBER].nunique()
        mvt_pct = self._get_votes(daos=df)[self.__DF_VOTER].nunique() / members[self.__DF_MEMBER].nunique()
        creation_date = df['summoningTime'].min()

        return Platform(
            name = 'DAOhaus',
            creation_date=creation_date,
            networks=list(df['network'].unique()),
            participation_stats = [
                MembersCreatedProposalsStat(mcp_pct),
                MembersEverVotedStat(mvt_pct),
            ],
        )

    def get_organization_list(self) -> OrganizationList:
        df = self._get_daos()
        members = self._get_members()

        # Getting the participation
        gbp = self._get_participation().groupby(self.__DF_IDX)
        gvt = self._get_votes().groupby(self.__DF_IDX)
        gbm = members.groupby(self.__DF_IDX)
        df['mcp_pct'] = gbp[self.__DF_PROPOSER].nunique() / gbm[self.__DF_MEMBER].nunique()
        df['mvt_pct'] = gvt[self.__DF_VOTER].nunique() / gbm[self.__DF_MEMBER].nunique()

        # Getting the activity
        dfgb = self._get_activity().groupby(self.__DF_IDX)[self.__DF_DATE]
        df['first_activity'] = dfgb.min()
        df['last_activity'] = dfgb.max()

        l: OrganizationList = OrganizationList()

        for _, org in df.iterrows():
            l.append(Organization(
                network = org['network'],
                o_id = org['molochAddress'],
                name = org['name'],
                creation_date = org['summoningTime'],
                first_activity = self._NaTtoNone(org['first_activity']),
                last_activity = self._NaTtoNone(org['last_activity']),
                participation_stats = [
                    MembersCreatedProposalsStat(org['mcp_pct']),
                    MembersEverVotedStat(org['mvt_pct']),
                ],
            ))

        return l



from typing import Dict
from dao_analyzer.apps.common.business.transfers import HierarchicalData, OrganizationList

from dao_analyzer.apps.common.business.i_metric_adapter import IMetricAdapter
import dao_analyzer.apps.aragon.data_access.daos.metric.\
    metric_dao_factory as s_factory

class AssetsValues(IMetricAdapter):
    HOVER_TEMPLATE = """%{label}<br>%{customdata[0]} $ <br>%{customdata[1]} €<br> %{customdata[2]} ETH<extra></extra>"""
    TEXT_TEMPLATE = """%{customdata[3]} %{label} <br> ~%{customdata[0]} $"""

    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param.
        """
        dao = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=s_factory.ASSETS_VALUES
        )

        metric: HierarchicalData = dao.get_metric()

        return {
            **metric.to_dict(),
            'root_color': 'lightgrey',
            'hovertemplate': self.HOVER_TEMPLATE,
            'texttemplate': self.TEXT_TEMPLATE,
            'branchvalues': 'remainder'
        }


"""
   Descp: This file is used to store the DAOhaus app's text.

   Created on: 1-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

TEXT: dict = {
    'approved_proposals': 'Approved proposals',
    'css_pane_border': 'daohaus-border',
    'css_id_assets': 'daohaus-assets',
    'css_id_activity': 'daohaus-activity',
    'css_id_member': 'daohaus-members',
    'css_id_organization': 'daohaus-organization',
    'css_id_rage_quit': 'daohaus-rage-quit',
    'css_id_vote': 'daohaus-vote',
    'css_id_proposal': 'daohaus-proposal',
    'disclaimer_activity': 'As activity we consider new proposals and votes',
    'dp_id_active_members': 'daohaus-dp-active-members',
    'dp_id_members': 'daohaus-dp-members',
    'dp_id_proposals': 'daohaus-dp-proposals',
    'dp_id_treasury': 'daohaus-dp-treasury',
    'dp_id_vote': 'daohaus-dp-vote',
    'dp_title_active_members': 'Active members',
    'dp_title_members': 'Members',
    'dp_title_proposals': 'New Proposals',
    'dp_title_treasury': 'Treasury',
    'dp_title_vote': 'Votes',
    'pane_css_prefix': 'daohaus-pane',
    'proposal_donation': 'Donation proposals',
    'proposal_grant': 'Grant proposals',
    'proposal_new_member': 'New member proposals',
    'proposal_other': 'Others',
    'rejected_proposals': 'Rejected proposals',
    'title_active_members': 'Active members',
    'title_active_organization': 'Number of DAOs active per month',
    'title_active_voters': 'Active voters',
    'title_activity': 'Activity',
    'title_approval_proposal_rate': 'Approval proposal rate',
    'title_assets': 'Assets',
    'title_assets_value': 'Tokens with fiat equivalence (USD)',
    'title_assets_novalue': 'Tokens without fiat equivalence',
    'title_member': 'Members',
    'title_new_members': 'New members',
    'title_new_proposals': 'New proposals',
    'title_organization_activity': 'Total activity per month',
    'title_out_members': 'Outgoing members',
    'title_proposal': 'Proposals',
    'title_proposal_outcomes': 'Proposal outcomes',
    'title_proposal_type': 'Proposals by type',
    'title_rage_quits': 'Rage Quit',
    'title_total_members': 'Total members',
    'title_vote': 'Votes',
    'title_vote_for_rate': 'Votes-for rate',
    'title_vote_against_rate': 'Votes-against rate',
    'title_vote_type': 'Votes for and against',
    'title_voters_percentage': 'Percentage of members who vote',
    'title_votes_voters': 'Number of votes per voter',
    'votes_against': 'Votes against',
    'votes_for': 'Votes for',
}



#!/usr/bin/env python3
import json
import sys
import pandas as pd
from pandas.testing import assert_frame_equal
from pathlib import Path

SKIP_COLUMNS = ['usdValue', 'eurValue', 'ethValue', 'balanceFloat']

def dfDiff(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
    merge = df1.merge(df2, how='outer', indicator=True)
    return merge[merge['_merge'] == 'left_only'].drop(columns="_merge")

def dfRead(f):
    df = pd.read_feather(f).set_index(['network', 'id'])
    df = df.sort_index().sort_index(axis=1) # Sorting columns and index
    df = df.drop(columns=SKIP_COLUMNS, errors='ignore')
    return df

def md_diff(d1: Path, d2: Path) -> bool:
    errors = False
    for f_left in d1.rglob('metadata.json'):
        f_right = d2 / f_left.relative_to(d1)

        with open(f_left, 'r') as f:
            j_left = json.load(f)
        with open(f_right, 'r') as f:
            j_right = json.load(f)

        if j_left['errors']:
            print(f'Error found in {f_left}: {j_left["errors"]}')
            errors = True
        if j_right['errors']:
            print(f'Error found in {f_right}: {j_right["errors"]}')
            errors = True
        
        for cid, cdata in j_left['metadata'].items():
            lb = cdata['block']
            rb = j_right['metadata'][cid]['block']
            if lb['id'] != rb['id']:
                errors = True
                print("> Metadata block mismatch")
                print(f"LEFT:  {lb}")
                print(f"RIGHT: {rb}, number: ")

    return errors

def dw_diff(d1: Path, d2: Path) -> bool:
    errors = False
    for f_left in d1.rglob('*.arr'):
        f_right = (d2 / f_left.relative_to(d1)).with_suffix('.arr')
        print(f"Checking {f_left} vs. {f_right}.", end='', flush=True)

        if not f_right.exists():
            errors = True
            print(f"Error with {f_right}:")
            print("> File doesn't exist yet")
            continue

        print(".", end='', flush=True)
        
        df_left = dfRead(f_left)
        df_right = dfRead(f_right)
        
        c_left = df_left.columns
        c_right = df_right.columns
        
        if not c_right.symmetric_difference(c_left).empty:
            errors = True
            print(f"\nError with {f_right}:")
            print("> NEW has " + ", ".join(c_right.difference(c_left)))
            print("> OLD has " + ", ".join(c_left.difference(c_right)))
            print("> Common  " + "," .join(c_right.intersection(c_left)))
            
            continue

        print(".", end='', flush=True)

        if (df_left.shape != df_right.shape):
            errors = True
            print(f'\nDifferent shape with {f_right}:')
            print("> LEFT  has", df_left.shape)
            print("> RIGHT has", df_right.shape)

        print(".", end='', flush=True)

        # dfc = df_left.compare(df_right)
        # if not dfc.empty:
        #     errors = True
        #     print(dfc)
        #     continue

        print(".", end='', flush=True)

        merge = df_left.merge(df_right, how='outer', indicator=True)
        left_merge = merge[merge['_merge'] == 'left_only']
        if not left_merge.empty:
            errors = True
            print(f"\nErrors with {f_left}:")
            print("LEFT ONLY (missing on right)")
            print(left_merge)

        right_merge = merge[merge['_merge'] == 'right_only']
        if not right_merge.empty:
            print(f"\nWarning with {f_right}:")
            print("RIGHT ONLY (missing on left)")
            print(right_merge)
        
        if not left_merge.empty or not right_merge.empty:
            continue
        
        print(".", end='', flush=True)

        try:
            assert_frame_equal(df_left, df_right, check_like=True)
        except AssertionError as e:
            errors = True
            print(f" Error with {f_right}:")
            print(e)
            continue
    
        print(" OK")

    return errors

if __name__ == "__main__":
    if len(sys.argv) != 3:
        print(f"Usage: {sys.argv[0]} datawarehouse1 datawarehouse2", file=sys.stderr)
        exit(2)

    d1 = Path(sys.argv[1])
    d2 = Path(sys.argv[2])
    if md_diff(d1, d2):
        exit(1)
    else:
        print("Metadata OK...")
    
    if dw_diff(d1, d2):
        exit(1)
            
    print("All OK!")
    exit(0)


"""
   Descp: Votes type metric strategy test.

   Created on: 8-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.apps.daohaus.data_access.daos.metric.strategy.\
    st_votes_type import StVotesType

from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie


class StVotesTypeTest(unittest.TestCase):

    def __check_lists(self, df: pd.DataFrame, out: List[List[int]]) -> None:
        strategy: StVotesType = StVotesType()
        result: StackedSerie = strategy.process_data(df=df)

        self.assertListEqual(out[0], result.get_i_stack(i_stack=0)) # test negatives
        self.assertListEqual(out[1], result.get_i_stack(i_stack=1)) # test positives


    def test_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=1, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'uintVote': 1, 'createdAt': bl.unix()},#today_year-(today_month-3)-01T00:00:00+00:00
            {'uintVote': 1, 'createdAt': bl.add(month=1).change(day=25).unix()},#today_year-(today_month-2)-25T00:00:00+00:00
            {'uintVote': 2, 'createdAt': bl.unix()},#today_year-(today_month-2)-25T00:00:00+00:00
            {'uintVote': 2, 'createdAt': bl.add(month=1).change(day=1, hour=23).unix()},#today_year-(today_month-1)-01T23:00:00+00:00
            {'uintVote': 1, 'createdAt': bl.sub(month=1).change(minute=59, second=59).unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'uintVote': 1, 'createdAt': bl.unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'uintVote': 1, 'createdAt': bl.add(month=2, second=1).change(day=21).unix()},#today_year-today_month-21T00:00:00+00:00
        ])

        out: List[List[int]] = [
            [0, 1, 1, 0], # number of votes against
            [1, 3, 0, 1]# number of votes for
        ]

        self.__check_lists(df=in_df, out=out)


    def test_process_data_votes_against(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=1, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'uintVote': 2, 'createdAt': bl.unix()},#today_year-(today_month-3)-01T00:00:00+00:00
            {'uintVote': 2, 'createdAt': bl.add(month=1).change(day=11).unix()},#today_year-(today_month-2)-11T00:00:00+00:00
            {'uintVote': 2, 'createdAt': bl.unix()},#today_year-(today_month-2)-11T00:00:00+00:00
            {'uintVote': 2, 'createdAt': bl.sub(month=1).change(day=3, hour=5).unix()},#today_year-(today_month-3)-03T5:00:00+00:00
            {'uintVote': 2, 'createdAt': bl.add(month=1).change(day=1, minute=59, second=59).unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'uintVote': 2, 'createdAt': bl.unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'uintVote': 2, 'createdAt': bl.change(day=21, hour=0, minute=0, second=0).unix()},#today_year-(today_month-2)-21T00:00:00+00:00
        ])

        out: List[List[int]] = [
            [2, 5, 0, 0], # number of votes against
            [0, 0, 0, 0]# number of votes for
        ]

        self.__check_lists(df=in_df, out=out)

if __name__ == "__main__":
    unittest.main()



"""
    Descp: common pieces of the cache_scripts entry point

    Created on: 16-feb-2022

    Copyright 2022 David Davó
        <david@ddavo.me>
"""

from .common import Collector, Runner, ENDPOINTS

__all__ = [
    'Collector',
    'Runner',
    'ENDPOINTS',
]


"""
   Descp: Class to adapt StakedSeries in a chart. This class is used to adapt
          the 'proposal outcome' metric

   Created on: 9-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict

import dao_analyzer.apps.common.resources.colors as Color
from dao_analyzer.apps.daohaus.resources.strings import TEXT
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.organization import OrganizationList
from dao_analyzer.apps.common.business.i_metric_adapter import IMetricAdapter
import dao_analyzer.apps.daohaus.data_access.daos.metric.\
    metric_dao_factory as s_factory

class ProposalOutcome(IMetricAdapter):

    DATE_FORMAT: str = '%b, %Y'

    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param.
        """
        dao = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=s_factory.PROPOSALS_OUTCOME
        )
        metric: StackedSerie = dao.get_metric()

        last_value: int = metric.get_last_value(0) + metric.get_last_value(1)

        return {
            'type1': {
                'y': metric.get_i_stack(0),
                'color': Color.LIGHT_RED,
                'name': TEXT['rejected_proposals'],
            },
            'type2': {
                'y': metric.get_i_stack(1),
                'color': Color.LIGHT_GREEN,
                'name': TEXT['approved_proposals'],
            },
            'common': {
                'x': metric.get_serie(),
                'type': 'date',
                'x_format': self.DATE_FORMAT,
                'ordered_keys': ['type1', 'type2'],
            },
            'last_serie_elem': metric.get_last_serie_elem(),
            'last_value': last_value,
            'diff': metric.get_diff_last_values(add_stacks=True),
            'diff_rel': metric.get_rel_last_values(add_stacks=True),
        }



import dao_analyzer.apps.common.resources.colors as Color

class LayoutConfiguration:
    def __init__(self) -> None:
        self.__color: str = Color.DARK_BLUE
        self.__css_border: str = ''
    
    @property
    def css_border(self) -> str:
        return self.__css_border

    @property
    def color(self) -> str:
        return self.__color

    def set_color(self, color: str) -> None:
        if color:
            self.__color = color

    def set_css_border(self, css_border: str) -> None:
        self.__css_border = css_border


"""
   Descp: Organization transfers.

   Created on: 3-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List, Any

from datetime import datetime
from .participation_stats import ParticipationStat

from dao_analyzer.apps.common.resources.strings import TEXT

class Organization:
    def __init__(self,
        o_id: str = TEXT['no_data'],
        name: str = TEXT['no_data'],
        network: str = TEXT['no_data'],
        creation_date: datetime = None,
        first_activity: datetime = None,
        last_activity: datetime = None,
        participation_stats: List[ParticipationStat] = [],
    ):
        self._id: str = o_id
        self._name: str = name
        self._network: str = network
        self._creation_date: datetime = creation_date
        self._first_activity: datetime = first_activity
        self._last_activity: datetime = last_activity
        self._participation_stats: List[ParticipationStat] = participation_stats

    # Redefinition of sorting functions
    def __eq__(self, other) -> bool:
        return self._id == other._id and self._network == other._network

    """ First we have the items with a name sorted by name. Then the ones without
        name sorted by id
    """
    def __lt__(self, other) -> bool:
        # If self doesn't have a name, we check if other has a name
        if self._name is None:
            if other._name is None:
                return self._id < other._id
            else:
                return False

        # If self has a name, but other does not have a name,
        # then self is lower than the other
        if other._name is None:
            return True
        
        # If the name is the same, we sort by network, then by id
        if self._name.casefold() == other._name.casefold():
            if self._network.casefold() == other._network.casefold():
                return self._id < other._id
            else:
                return self._network.casefold() < other._network.casefold()
        else:
            return self._name.casefold() < other._name.casefold()
    
    def get_dict_representation(self) -> Dict[str, str]:
        return {
            'value': self._id, 
            'label': self.get_label()
        }

    def to_plotly_json(self) -> Dict[str, Any]:
        """ Allows to store the object inside a Dcc.Store """
        return {
            'address': self.get_id(),
            'name': self.get_name(),
            'network': self.get_network(),
            'creation_date': self.get_creation_date(),
            'first_activity': self.get_first_activity(),
            'last_activity': self.get_last_activity(),
            'participation_stats': self.get_participation_stats(),
        }

    @classmethod
    def from_json(cls, dict: Dict[str, Any]) -> 'Organization':
        def _getdt(key):
            d = dict.get(key, None)

            if not d:
                return None
            elif isinstance(d, str):
                return datetime.fromisoformat(d)
            else:
                raise TypeError

        participation = [ParticipationStat.from_json(x) for x in dict.get('participation_stats', [])]

        return Organization(
            o_id = dict['address'],
            network = dict['network'],
            name = dict.get('name', None),
            creation_date = _getdt('creation_date'),
            first_activity = _getdt('first_activity'),
            last_activity = _getdt('last_activity'),
            participation_stats = participation,
        )

    def get_id(self) -> str:
        return self._id

    def get_name(self) -> str:
        return self._name

    def get_network(self) -> str:
        return self._network

    def get_creation_date(self) -> datetime:
        return self._creation_date

    def get_first_activity(self) -> datetime:
        return self._first_activity

    def get_participation_stats(self) -> List[ParticipationStat]:
        return self._participation_stats

    def get_last_activity(self) -> datetime:
        return self._last_activity

    def get_label(self) -> str:
        if not self._name:
            return f"{self._id[:16]}... ({self._network})"
        
        return f"{self._name} ({self._id[:8]}... {self._network})"

    def get_label_long(self) -> str:
        if not self._name:
            return f"{self._id} ({self._network})"

        return f"{self._name} — {self._id} ({self._network})"



"""
   Descp: Common pane which wraps data table and other components.

   Created on: 15-mar-2022

   Copyright 2020-2022 David Davó Laviña
        <ddavo@ucm.es>
"""
from dash import html, dcc
from dash.dash_table import DataTable

from . import ILayout
from .data_table_configuration import DataTableConfiguration

class DataTableLayout(ILayout):
    SUFFIX_ID_TABLE: str = '-dtable'
    
    def __init__(self, title: str, css_id: str) -> None:
        self.__title: str = title
        self.__css_id: str = css_id
        self.__configuration: DataTableConfiguration = DataTableConfiguration()
    
    @property
    def table_id(self):
        return f'{self.__css_id}{self.SUFFIX_ID_TABLE}'

    @property
    def configuration(self) -> DataTableConfiguration:
        return self.__configuration

    def get_layout(self) -> html.Div:
        children = [
            html.Div(children=[
                html.Span(
                    self.__title,
                    className='graph-pane-title'
                ),
                html.Span('nothing', className='hide'),
            ], className='d-flex flex-column chart-header'),
            DataTable(
                id=self.table_id,
                # style_table={'height': '450px', 'margin-top': '40px', 'overflowY': 'auto'},
                sort_action='native',
                page_size=14
            )
        ]

        return html.Div(
            children=dcc.Loading(
                type="circle",
                color=self.configuration.color,
                children=html.Div(
                    children=children,
                    id=self.__css_id,
                    className='d-flex flex-column'
                ),
                className='dcc-loading',
                parent_style={'flex': '1 1 auto'},
                style={'flex': '1 1 auto'}
            ),
            className=f'pane {self.configuration.css_border} col',
            style={'align-self': 'stretch', 'display': 'flex'}
        )


"""
   Descp: It's used to create the dashboard view.

   Created on: 20-feb-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import os
from typing import Dict, List, Callable
from dash import dcc, html
import dash_bootstrap_components as dbc
from dao_analyzer.apps.common.business.transfers.organization import Organization, OrganizationList
from dao_analyzer.apps.common.business.transfers.organization.organization_filter import NetworkRadioButton, OrganizationFilterGroup
from dao_analyzer.apps.common.business.transfers.organization.participation_stats import ParticipationStat
from dao_analyzer.apps.common.business.transfers.organization.platform import Platform

from dao_analyzer.apps.common.resources import colors as COLOR
from dao_analyzer.apps.common.resources.strings import TEXT
from dao_analyzer.apps.common.presentation.main_view.main_view import REL_PATH

__ECOSYSTEM_SELECTED: Dict[str, List[str]] = {
    'default': ['', '', ''],
    'daohaus': ['daohaus-selected', '', ''],
    'aragon': ['', 'aragon-selected', ''],
    'daostack': ['', '', 'daostack-selected'],
}

def generate_layout(
    organization_list: OrganizationList, 
    sections: Dict,
    datapoints, 
    ecosystem: str, 
    update: str, 
    platform_id: str, 
    org_value: str,
    network_value: str,
) -> List:
    """
    Use this function to generate the app view.
    Params:
        labels: A list of dictionaries for each element, in order to 
        fill the dropdown selector.
    Return:
        A html.Div filled with the app view 
    """
    if not org_value:
        org_value = organization_list.ALL_ORGS_ID

    return html.Div([
        dbc.Container(
            __generate_header(organization_list, ecosystem, update, org_value, network_value),
        className='top body mb-3 py-4'),
        dbc.Container([
            __generate_subheader(platform_id, datapoints),
            __generate_sections(sections),
        ], className='body', id=f'{platform_id}-body'),
    ])

def __gen_ecosystem(id: str, selected: str) -> html.Div:
    return html.Div(children=[
        html.Div(className=f'ecosystem-overlay {id}-color',
            id=f'{id}-bt'),
        html.Img(src=os.path.join(REL_PATH, TEXT[f'{id}_image_name']),
            className='ecosystem-img'),
    ], className=f'ecosystem {id}-ecosystem {selected}')

def __generate_header(
    organization_list: OrganizationList,
    ecosystem: str,
    update: str,
    org_value: str,
    network_value: str,
) -> dbc.Row:
    selected: List[str] = __ECOSYSTEM_SELECTED['default']
    if ecosystem in __ECOSYSTEM_SELECTED.keys():
        selected = __ECOSYSTEM_SELECTED[ecosystem]

    ecosystems: List[html.Div] = [ __gen_ecosystem(eid, selected[i]) for i,eid in enumerate(['daohaus', 'aragon', 'daostack']) ]

    if not org_value:
        org_value = OrganizationList.ALL_ORGS_ID
    
    # Disable filters if generating header with a DAO pre-selected (comes from URL)
    # If we don't disable the filter, the DAO will be inmediately filtered out
    # Fixes #85
    filterGroup: OrganizationFilterGroup = organization_list.get_filter_group(
        force_disabled=not OrganizationList.is_all_orgs(org_value)
    )

    try:
        networkRadio: NetworkRadioButton = organization_list.get_network_radio(network_value)
    except ValueError:
        networkRadio: NetworkRadioButton = organization_list.get_network_radio()

    return dbc.Row(children=[
        # 1. Ecosystem selector
        html.Div(
            html.Div([
                html.Div(TEXT['ecosystem_selector_title']),
                html.Div(children=ecosystems, className='ecosystems-wrapper'),
            ], className='select-platform-wrapper'),
        className='col d-flex justify-content-center'),
        # 2. Network selector
        html.Div(
            html.Div(children=[
                html.Div(html.Span(TEXT['network_selector_title'])),
                dcc.RadioItems(
                    options = networkRadio.get_options(),
                    value = networkRadio.get_value(),
                    className = 'checklist-filter',
                    id='org-network-radio'),
            ], className='select-network-wrapper'),
        className='col-3 d-flex justify-content-center'),
        # 3. DAO selector (and filtering)
        html.Div(
            html.Div(children=[
                html.Div(html.Span(TEXT['dao_selector_title'])),
                html.Div([
                    dcc.Checklist(
                        options = filterGroup.get_options(),
                        value = filterGroup.get_values(),
                        className='checklist-filter d-flex flex-column',
                        id='org-filter',
                    ),
                    html.Div(dcc.Dropdown(
                        id='org-dropdown',
                        options=organization_list.get_dict_representation(),
                        value=org_value,
                        clearable=False,
                    )),
                    html.Div("", id='org-number', className='number-of-daos'),
                ], className='flex-grow-1'),
            ], className='select-dao-wrapper'),
        className='col d-flex justify-content-center'),
        # 4. information storing (hidden from user)
        # The following dcc is changed in callbacks that modify the dao-info-container
        dcc.Store(
            id='platform-info-store',
            data=None, # Initially there is no data
            storage_type='memory',
        ),
        # This dcc is read-only an used to generate the dropdown
        dcc.Store(
            id='organization-list-store',
            data=organization_list,
            storage_type='memory',
        ),
        html.Div(f'Last update: {update}', className='last-update'),
    ], className='body-header row-divider')

### SUBHEADER THINGS
def _gen_participation_stats(stats: List[ParticipationStat]) -> html.Div:
    children = []
    for s in stats:
        if s.value is None:
            children.append(html.Div(s.text))
        else:
            children.append(html.Div([html.B(s.value_str), s.text]))

    return html.Div(children, className='dao-info-stats')

def _get_platform_info(p: Platform) -> html.Div:
    if p is None:
        return html.Div(TEXT['no_data_selected'], className='dao-info-name')
    
    grid: List[html.Div] = [
        html.Div('Platform', className='dao-info-label'),
        html.Div(p.name, className='dao-info-name'),
        html.Div('Networks', className='dao-info-label'),
        html.Div(', '.join(p.networks), className='dao-info-network'),
    ]

    if p.creation_date:
        grid.append(html.Div("Creation Date", className='dao-info-label'))
        grid.append(html.Div(p.creation_date.strftime(TEXT['creation_date_format']), className='dao-info-date'))

    if p.participation_stats:
        grid.append(html.Div("Participation", className='dao-info-label'))
        grid.append(_gen_participation_stats(p.participation_stats))
    
    return html.Div(grid, className='dao-info-container')

def _get_dao_info(org: Organization) -> html.Div:
    name = html.I(TEXT['unknown_dao_name'])
    if org.get_name():
        name = org.get_name()

    grid: List[html.Div] = [
        html.Div("DAO", className='dao-info-label'),
        html.Div(name, className='dao-info-name'),
        html.Div("Network", className='dao-info-label'),
        html.Div(org.get_network(), className='dao-info-network'),
        html.Div("Address", className='dao-info-label'),
        html.Div(html.Span(org.get_id(), className='address'), className='dao-info-address'),
    ]

    if org.get_creation_date():
        grid.append(html.Div("Creation Date", className='dao-info-label'))
        grid.append(html.Div(org.get_creation_date().strftime(TEXT['creation_date_format']), className='dao-info-date'))
    elif org.get_first_activity():
        grid.append(html.Div("First Activity", className='dao-info-label'))
        grid.append(html.Div(org.get_first_activity().strftime(TEXT['creation_date_format']), className='dao-info-date'))

    if org.get_participation_stats():
        grid.append(html.Div("Participation", className='dao-info-label'))
        grid.append(_gen_participation_stats(org.get_participation_stats()))
    
    return html.Div(grid, className='dao-info-container')

def _gen_sum_hdr(org: Organization = None):
    if not org:
        return None
    
    left = html.Span()
    if org.get_last_activity():
        right = html.Span(['Last active on ', html.B(org.get_last_activity().strftime(TEXT['last_activity_format']))])
    else:
        right = html.Span(['This organization has ', html.B('never'), ' been active'])
    
    return [left, right]

def _get_dao_summary_layout(org_id, datapoints: Dict ):
    dp_divs: List[html.Div] = [ dcc.Loading(
        dp.get_layout(), 
        type='circle',
        color=COLOR.DARK_BLUE
    ) for dp in datapoints.values() ]

    return html.Div([
        html.Div(_gen_sum_hdr(), className='dao-summary-hdr', id=org_id+'-summary-hdr'),
        html.Div(dp_divs, className='dao-summary-body'),
    ], className='dao-summary-container')

def __generate_subheader(org_id: str, datapoints: Dict[str, List[Callable]]) -> dbc.Row:
    return dbc.Row(
        id=org_id,
        className='my-3',
        children=html.Div([
           dbc.Col(dcc.Loading(html.Div(_get_platform_info(None), id=org_id+'-info'), color=COLOR.DARK_BLUE)),
           dbc.Col(_get_dao_summary_layout(org_id, datapoints)),
        ], className='dao-header-container pt-4'),
    )

def __generate_sections(sections: Dict[str, List[Callable]]) -> dbc.Row:
    tabs: List[dcc.Tab] = []

    for name, data in sections.items():
        charts = list()
        for chart_pane in data['callables']:
            charts.append(chart_pane())

        sec_hdr = dbc.Row(
            [html.Div(name, className='section-title')],
            id=f'{data["css_id"]}-hdr', 
            className='section-hdr'
        )

        if 'disclaimer' in data and data['disclaimer']:
            sec_hdr.children.append(html.Div(data['disclaimer'], className='section-disclaimer'))

        container = dbc.Container(
            class_name='g-4 mt-4',
            children=[
                sec_hdr,
                dbc.Row(children=charts, className='row-cols-1 row-cols-xl-2 gx-5'),
            ],
        )

        tabs.append(dcc.Tab(label=name, children=container))

    return dcc.Tabs(tabs)



import logging
from typing import Callable, List, Dict
from abc import ABC, abstractmethod

import pandas as pd

from gql.dsl import DSLField

from .common import ENDPOINTS, NetworkCollector, NetworkRunner, Runner, UpdatableCollector
from .api_requester import GQLRequester
from ..metadata import Block

def add_where(d, **kwargs):
    """
    Adds the values specified in kwargs to the where inside d
        Example: `**add_where(kwargs, deleted=False)`
    """
    if "where" in d:
        d["where"] |= kwargs
    else:
        d["where"] = kwargs
    
    return d

def partial_query(q, w) -> DSLField:
    def wrapper(**kwargs):
        return q(**add_where(kwargs, **w))
    return wrapper

def checkSubgraphHealth(endpoint: str, network: str = None):
    subgraphName = '/'.join(endpoint.split('/')[-2:])
    
    requester = GQLRequester(endpoint=ENDPOINTS['_theGraph']['index-node'])
    ds = requester.get_schema()
    q = ds.Query.indexingStatusForCurrentVersion(subgraphName=subgraphName).select(
        ds.SubgraphIndexingStatus.health,
        ds.SubgraphIndexingStatus.synced,
        ds.SubgraphIndexingStatus.node,
        ds.SubgraphIndexingStatus.chains.select(
            ds.ChainIndexingStatus.network
        )
    )

    r = requester.request_single(q)

    if not r['synced']:
        logging.info(f"Subgraph {endpoint} is not synced")

    if r['health'].lower() != 'healthy':
        logging.error(f"Subgraph {endpoint} is not healthy")
        return False
    
    subgraph_network = r['chains'][0]['network']
    if network and subgraph_network != network:
        logging.error(f"Subgraph {endpoint} expected network {network} but got {subgraph_network}")
        return False

    return True

class GraphQLCollector(NetworkCollector, UpdatableCollector):
    def __init__(self, name: str, runner: Runner, endpoint: str, result_key: str = None, index: str = None, network: str='mainnet', pbar_enabled: bool=True):
        super().__init__(name, runner, network)
        self.endpoint: str = endpoint
        self.index = index if index else 'id'
        self.result_key = result_key if result_key else name
        self.postprocessors: Callable = []

        self.requester: GQLRequester = GQLRequester(endpoint=self.endpoint, pbar_enabled=pbar_enabled)

    def postprocessor(self, f: Callable[[pd.DataFrame], pd.DataFrame]):
        self.postprocessors.append(f)
        return f

    @property
    def schema(self):
        return self.requester.get_schema()

    @abstractmethod
    def query(self, **kwargs) -> DSLField:
        raise NotImplementedError

    @property
    def df(self) -> pd.DataFrame:
        if not self.data_path.is_file():
            return pd.DataFrame()

        df = pd.read_feather(self.data_path)
        if self.network:
            df = df[df['network'] == self.network]
        
        return df

    def transform_to_df(self, data: List[Dict], skip_post: bool=False) -> pd.DataFrame:
        df = pd.DataFrame.from_dict(pd.json_normalize(data))

        # For compatibility reasons we change from . to snake case
        def dotsToSnakeCase(str: str) -> str:
            splitted = str.split('.')
            return splitted[0] + ''.join(x[0].upper()+x[1:] for x in splitted[1:])
                        
        df = df.rename(columns=dotsToSnakeCase)
        df['network'] = self.network

        if not skip_post:
            for post in self.postprocessors:
                logging.debug(f"Running postprocessor {post.__name__}")
                df = post(df)
                if df is None:
                    raise ValueError(f"The postprocessor {post.__name__} returned None")

        return df

    def verify(self) -> bool:
        # Checking if the queryBuilder doesn't raise any errors
        self.query()

        # Checking the health of the subgraph
        return checkSubgraphHealth(self.endpoint)

    def query_cb(self, prev_block: Block = None):
        if prev_block:
            return partial_query(self.query, {'_change_block': {'number_gte': prev_block.number}})
        else:
            return self.query

    def run(self, force=False, block: Block = None, prev_block: Block = None):
        logging.info(f"Running GraphQLCollector with block: {block}, prev_block: {prev_block}")
        if block is None:
            block = Block()
        if prev_block is None or force:
            prev_block = Block()

        data = self.requester.n_requests(query=self.query_cb(prev_block), block_hash=block.id)

        # transform to df
        df: pd.DataFrame = self.transform_to_df(data)
        self._update_data(df, force)

class GraphQLRunner(NetworkRunner, ABC):
    pass


"""
   Descp: This class is used to adapt the proposal majority
          metric to its visual representation.

   Created on: 14-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List
from dao_analyzer.apps.common.business.transfers.organization.organization_list import OrganizationList

import dao_analyzer.apps.common.resources.colors as Color
from dao_analyzer.apps.daostack.resources.strings import TEXT
from dao_analyzer.apps.daostack.business.metric_adapter.metric_adapter import MetricAdapter
from dao_analyzer.apps.common.business.transfers.n_stacked_serie import NStackedSerie
import dao_analyzer.apps.daostack.data_access.daos.metric.\
    metric_dao_factory as s_factory

class MajorityType(MetricAdapter):
    def __init__(self, metric_id: int) -> None:
        super().__init__(metric_id)

    
    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param.
        """
        dao = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=super().metric_id
        )
        metric: NStackedSerie = dao.get_metric()

        y1 = metric.get_i_sserie(0)
        y2 = metric.get_i_sserie(1)
        y3 = metric.get_i_sserie(2)
        y4 = metric.get_i_sserie(3)
        x: List = y1.get_serie()

        return {
            'type1': {
                'x': y1.get_serie(),
                'y': y1.get_i_stack(0),
                'color': f'rgba{Color.hex_to_rgba(Color.DARK_GREEN, 0.5)}',
                'marker_symbol': 'triangle-up',
                'name': TEXT['abs_pass'],
                'position': 'up',
            },
            'type2': {
                'x': y2.get_serie(),
                'y': y2.get_i_stack(0),
                'color': f'rgba{Color.hex_to_rgba(Color.DARK_GREEN, 0.5)}',
                'marker_symbol': 'circle',
                'name': TEXT['rel_pass'],
                'position': 'up',
            },
            'type3': {
                'x': y3.get_serie(),
                'y': y3.get_i_stack(0),
                'color': f'rgba{Color.hex_to_rgba(Color.DARK_RED, 0.5)}',
                'marker_symbol': 'circle',
                'name': TEXT['rel_fail'],
                'position': 'down',
            },
            'type4': {
                'x': y4.get_serie(),
                'y': y4.get_i_stack(0),
                'color': f'rgba{Color.hex_to_rgba(Color.DARK_RED, 0.5)}',
                'marker_symbol': 'triangle-down',
                'name': TEXT['abs_fail'],
                'position': 'down',
            },
            'common': {
                'x': x,
                'type': 'date', 
                'x_format': super().DATE_FORMAT,
                'ordered_keys': ['type1', 'type2', 'type3', 'type4'], 
                'y_suffix': '%',
            }
        }



"""
   Descp: Strategy pattern for proposal's boost outcomes in a timeline 

   Created on: 13-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List, Tuple, Any
import pandas as pd

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.n_stacked_serie import NStackedSerie 
from dao_analyzer.apps.common.business.transfers.serie import Serie
import dao_analyzer.apps.common.data_access.pandas_utils as pd_utl


BOOST_OUTCOME: int = 0
BOOST_SUCCESS_RATIO: int = 1
TOTAL_SUCCESS_RATIO: int = 2


class StProposalOutcome(IMetricStrategy):
    __DF_DATE = 'executedAt'
    __DF_PASS = 'hasPassed'
    __DF_BOOST = 'isBoosted'
    __DF_BOOST_AT = 'boostedAt'
    __DF_OUTCOME = 'winningOutcome'
    __DF_REP = 'totalRepWhenExecuted'
    __DF_VOTES_F = 'votesFor'
    __DF_QUORUM = 'queuedVoteRequiredPercentage'
    __DF_COUNT = 'count'
    __DF_COLS1 = [__DF_DATE, __DF_PASS, __DF_BOOST]
    __DF_COLS2 = [__DF_DATE, __DF_PASS, __DF_BOOST, __DF_COUNT]
    __DF_INI_COLS = [__DF_DATE, __DF_OUTCOME, __DF_REP, __DF_VOTES_F,
                    __DF_BOOST_AT, __DF_QUORUM]


    def __init__(self, m_type: int):
        self.__m_type = m_type


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff = dff.dropna(subset=[self.__DF_DATE])
        dff.loc[:, self.__DF_INI_COLS] = dff[self.__DF_INI_COLS]
        return dff


    def __get_boost_from_dataframe(self, df: pd.DataFrame, boosted: bool)\
    -> Tuple[List[int]]:
        s_pass: List[int] = list()
        s_not_pass: List[int] = list()

        boost: List[int] = df[self.__DF_BOOST].tolist()
        is_pass: List[int] = df[self.__DF_PASS].tolist()
        count: List[int] = df[self.__DF_COUNT].tolist()

        for i, b in enumerate(boost):
            if b == boosted:
                if is_pass[i]:
                    s_pass.append(count[i])
                else:
                    s_not_pass.append(count[i])

        return (s_not_pass, s_pass)


    def __get_empty_transfer(self) -> Any:
        if self.__m_type == BOOST_OUTCOME:
            return StackedSerie()
        elif self.__m_type == TOTAL_SUCCESS_RATIO:
            return StackedSerie()
        elif self.__m_type == BOOST_SUCCESS_RATIO:
            return NStackedSerie()

        return None


    def process_data(self, df: pd.DataFrame) -> Any:
        if pd_utl.is_an_empty_df(df):
            return self.__get_empty_transfer()

        df = self.clean_df(df=df)
        df = self.__calculate_outcome(df=df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        df = pd_utl.count_cols_repetitions(df=df, 
            cols=self.__DF_COLS1, new_col=self.__DF_COUNT)

        # generates a time serie
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)

        # joinning all the data in a unique dataframe and fill with all combinations
        for p in [True, False]:
            for b in [True, False]:
                dff = pd_utl.get_df_from_lists([idx, p, b, 0], self.__DF_COLS2)

                dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)
                df = pd.concat([df, dff], ignore_index=True)

        df = df.drop_duplicates(subset=self.__DF_COLS1, keep="first")
        df = df.sort_values(self.__DF_DATE, ignore_index=True)

        return self.generate_metric(df)


    def generate_metric(self, df: pd.DataFrame) -> Any:
        metric = StackedSerie()

        if self.__m_type == BOOST_OUTCOME:
            metric = self.__get_boost_outcome(df)
        elif self.__m_type == TOTAL_SUCCESS_RATIO:
            metric = self.__get_total_success_ratio(df)
        elif self.__m_type == BOOST_SUCCESS_RATIO:
            metric = self.__get_boost_success_ratio(df)

        return metric


    def __get_boost_outcome(self, df: pd.DataFrame) -> StackedSerie:
        serie: Serie = Serie(x = df.drop_duplicates(subset=self.__DF_DATE,
            keep="first")[self.__DF_DATE].tolist())

        n_p1, p1 = self.__get_boost_from_dataframe(df, boosted=False)
        n_p2, p2 = self.__get_boost_from_dataframe(df, boosted=True)

        return StackedSerie(serie=serie, y_stack=[p1, p2, n_p2, n_p1])


    def __get_total_success_ratio(self, df: pd.DataFrame) -> StackedSerie:
        serie: Serie = Serie(x = df.drop_duplicates(subset=self.__DF_DATE,
            keep="first")[self.__DF_DATE].tolist())

        tp: List[int] = self.__get_predicted_values(df, 'tp')
        tn: List[int] = self.__get_predicted_values(df, 'tn')
        fp: List[int] = self.__get_predicted_values(df, 'fp')
        fn: List[int] = self.__get_predicted_values(df, 'fn')

        ratio: List[float] = self.__calculate_ratio(
            numerator=[tp, tn], 
            denominator=[tp, tn, fp, fn], 
            _len=len(tp))

        return StackedSerie(serie=serie, y_stack=[ratio])


    def __get_boost_success_ratio(self, df: pd.DataFrame) -> NStackedSerie:
        serie: Serie = Serie(x = df.drop_duplicates(subset=self.__DF_DATE,
            keep="first")[self.__DF_DATE].tolist())

        tp: List[int] = self.__get_predicted_values(df, 'tp')
        tn: List[int] = self.__get_predicted_values(df, 'tn')
        fp: List[int] = self.__get_predicted_values(df, 'fp')
        fn: List[int] = self.__get_predicted_values(df, 'fn')

        boost_ratio: List[float] = self.__calculate_ratio(
            numerator=[tp], 
            denominator=[tp, fp], 
            _len=len(tp))

        nboost_ratio: List[float] = self.__calculate_ratio(
            numerator=[tn], 
            denominator=[tn, fn], 
            _len=len(tn))

        return NStackedSerie(
            serie=serie, 
            sseries=[
                StackedSerie(y_stack=[boost_ratio]),
                StackedSerie(y_stack=[nboost_ratio])])


    def __calculate_ratio(self, numerator: List, denominator: List, 
    _len: int) -> List:

        ratio: List = list()
        for i in range(_len):
            n_val: int = 0
            d_val: int = 0

            # numerator elements sum
            for n in numerator:
                n_val += n[i]

            # denominator elements sum
            for d in denominator:
                d_val += d[i]

            if d_val == 0:
                ratio.append(None)
            else:
                ratio.append(round(n_val / d_val, 4))

        return ratio

    
    def __calculate_outcome(self, df: pd.DataFrame) -> pd.DataFrame:
        df[self.__DF_BOOST] = ~pd.isna(df[self.__DF_BOOST_AT])
        df[self.__DF_PASS] = df.apply(lambda r: self.__has_passed(r, r[self.__DF_BOOST]), axis='columns')
        df = df.filter(self.__DF_COLS1).astype(object)
        return df


    def __has_passed(self, data, is_boost: bool) -> bool:
        # winning outcome means more votes for than votes against
        outcome: bool = data[self.__DF_OUTCOME] == 'Pass'
        percentage = int(data[self.__DF_VOTES_F]) / int(data[self.__DF_REP]) * 100 \
            if int(data[self.__DF_REP]) > 0 else 0
        limit: int = int(data[self.__DF_QUORUM])

        has_passed: bool = outcome and is_boost
        # if there was no boost and pass outcome, you must consider if there was absolute majority
        if outcome and not is_boost:
            has_passed = percentage >= limit

        return has_passed


    def __get_predicted_values(self, df: pd.DataFrame, pred: str) -> List[int]:
        """ 
        True positives = boost and pass
        True negatives = not boost and not pass
        False positives = boost and not pass
        False negatives = not boost and pass

        Parameters:
            * df = data frame to filter
            * pred = must be 'tp', 'tn', 'fp', 'fn' in other case return true 
                     positive by default.
        Return:
            A list with a counter of predicted values. This list has the 
            number of elements such as different dates of the df parameter.
            This list is also ordered by the dates of the df parameter. 
        """
        # default tp
        boost: bool = True
        _pass: bool = True

        if pred == 'tn':
            boost = False
            _pass = False
        elif pred == 'fp':
            _pass = False
        elif pred == 'fn':
            boost = False

        dff = pd_utl.filter_by_col_value(df, self.__DF_BOOST, boost, [pd_utl.EQ])
        dff = pd_utl.filter_by_col_value(dff, self.__DF_PASS, _pass, [pd_utl.EQ])

        # date reconstruction
        idx = pd_utl.get_monthly_serie_from_df(dff, self.__DF_DATE)
        d3f = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        d3f = pd_utl.datetime_to_date(d3f, self.__DF_DATE)

        dff = pd_utl.filter_by_col_value(dff, self.__DF_COUNT, 0, [pd_utl.GT])
        dff = dff.drop(columns=[self.__DF_BOOST, self.__DF_PASS])

        dff = pd.concat([dff, d3f], ignore_index=True)
        dff = dff.drop_duplicates(subset=self.__DF_DATE, keep="first")
        dff = dff.sort_values(self.__DF_DATE, ignore_index=True)

        return dff[self.__DF_COUNT].to_list()
    


#!/bin/env python
import requests
import json
import re
import sys
from pathlib import Path

if len(sys.argv) != 2:
    print(f"Usage: {sys.argv[0]} filename", file=sys.stderr)
    exit(1)

PREV_FILE = Path(sys.argv[1])

# flake8: noqa W605
KNOWN_ORGS_FILE_URL="https://github.com/aragon/client/raw/develop/src/known-organizations/index.js"
REGEX_MAIN_DICT = re.compile("export\s+const\s+KnownOrganizations\s+=\s+{\s+main:[^\[]+\[([^\]]*)\]", re.MULTILINE)
REGEX_ITEM_DICT = {
    "address": re.compile("address:\s+'(0x[0123456789abcdefABCDEF]{0,40})'"),
    "domain": re.compile("domain:\s+'([\.\w]*)'"),
    "name": re.compile("name:\s+'([\w\s\.]+)'"),
} 
r = requests.get(KNOWN_ORGS_FILE_URL)

assert(r.status_code == 200)

extracted_list = re.findall("(\{[^\}]+\})," ,REGEX_MAIN_DICT.findall(r.text)[0], re.MULTILINE)

def _parse_item(item):
    return {k:v.findall(item)[0] for k,v in REGEX_ITEM_DICT.items()}

with open(PREV_FILE, 'r') as pf:
    prev_json = json.loads(pf.read())

processed_list = [_parse_item(x) for x in extracted_list]
processed_dict = {"mainnet": processed_list}

print(json.dumps(prev_json | processed_dict, indent=2))



"""
   Descp: Manage the application logic, and it's used to interconect the
        data_access and presentation layers.

   Created on: 2-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List, Callable
from dash import html

from dao_analyzer.apps.common.business.i_metric_adapter import IMetricAdapter
from dao_analyzer.apps.common.business.transfers import Platform, OrganizationList
from dao_analyzer.apps.common.presentation.charts.chart_sum_controller import ChartSummaryController
import dao_analyzer.apps.common.presentation.dashboard_view.dashboard_view as view
import dao_analyzer.apps.common.presentation.dashboard_view.controller as view_cont
from dao_analyzer.apps.daostack.data_access.daos.platform_dao import DaostackDAO
from dao_analyzer.apps.common.data_access.daos.platform_dao import PlatformDAO
from dao_analyzer.apps.common.presentation.data_point_layout import DataPointLayout
import dao_analyzer.apps.daostack.data_access.daos.metric.\
    metric_dao_factory as s_factory
from dao_analyzer.apps.common.business.singleton import Singleton
from dao_analyzer.apps.common.presentation.charts.chart_controller import ChartController
from dao_analyzer.apps.daostack.business.metric_adapter.metric_adapter import MetricAdapter
from dao_analyzer.apps.daostack.business.metric_adapter.proposal_boost_outcome \
    import ProposalBoostOutcome
from dao_analyzer.apps.daostack.business.metric_adapter.success_ratio_type \
    import SuccessRatioType
from dao_analyzer.apps.daostack.business.metric_adapter.vote_type \
    import VoteType
from dao_analyzer.apps.daostack.business.metric_adapter.majority_type \
    import MajorityType
from dao_analyzer.apps.daostack.business.metric_adapter.asset_tokens import AssetsTokens
from dao_analyzer.apps.daostack.business.metric_adapter.asset_values import AssetsValues
from dao_analyzer.apps.common.presentation.charts.dt_controller import DataTableController
from dao_analyzer.apps.common.presentation.charts.layout import ChartPaneLayout, DataTableLayout
from dao_analyzer.apps.common.presentation.charts.layout.figure import BarFigure, CalFigure, MultiBarFigure, DoubleScatterFigure, Figure, TreemapFigure
from dao_analyzer.apps.daostack.resources.strings import TEXT

class DaostackService(metaclass=Singleton):
 
    _REP_H: int = 0
    _VOTE: int = 1
    _STAKE: int = 2
    _PROPOSAL: int = 3
    _ORGANIZATION: int = 4
    _ASSETS: int = 5

    def __init__(self):
        # app state
        self.__orgsDAO: PlatformDAO = DaostackDAO()
        self.__controllers: Dict[int, List[ChartController]] = {
            self._REP_H: list(),
            self._VOTE: list(),
            self._STAKE: list(),
            self._PROPOSAL: list(),
            self._ORGANIZATION: list(),
            self._ASSETS: list(),
        }
        self.__already_bound: bool = False
        self.__data_points: Dict[str, DataPointLayout] = {}


    def bind_callbacks(self, app) -> None:
        if not self.__already_bound:
            self.__already_bound = True
            # Changing the DAO name if it changes
            view_cont.bind_callbacks(
                app=app,
                section_id=TEXT['css_id_organization'],
            )
            self.__gen_sections()

            for contList in self.__controllers.values():
                for c in contList:
                    if hasattr(c, 'bind_callback'):
                        c.bind_callback(app)


    def platform(self, orglist = None) -> Platform:
        return self.__orgsDAO.get_platform(orglist)

    def organization_list(self) -> OrganizationList:
        return self.__orgsDAO.get_organization_list()

    @property
    def are_panes(self) -> bool:
        """
        Checks if panes and their controllers are already created.
        """
        return any(self.__controllers.values())


    # Called every request
    def get_layout(self, **kwargs) -> html.Div:
        """
        Returns the app's layout. 
        """
        if not self.__already_bound:
            self.bind_callbacks()

        return view.generate_layout(
            organization_list=self.organization_list(),
            sections=self.__get_sections(),
            ecosystem='daostack',
            update=self.__orgsDAO.get_last_update_str(),
            platform_id=TEXT['css_id_organization'],
            datapoints=self.__get_datapoints(),
            **kwargs,
        )

    def __gen_sections(self) -> None:
        self.__get_rep_holder_charts()
        self.__get_vote_charts()
        self.__get_stake_charts()
        self.__get_proposal_charts()
        self.__get_organization_charts()
        self.__get_assets_charts()

    def __get_sections(self) -> Dict[str, List[Callable]]:
        """
        Returns a dict with each section filled with a callable function, which
         returns the chart layout.
        """
        l_rep_h: List[Callable] = list()
        l_vote: List[Callable] = list()
        l_stake: List[Callable] = list()
        l_proposal: List[Callable] = list()
        l_organization: List[Callable] = list()
        l_assets: List[Callable] = list()

        if not self.are_panes:
            self.__gen_sections()

        # Panes are already created.
        l_rep_h = [c.layout.get_layout for c in self.__controllers[self._REP_H]]
        l_vote = [c.layout.get_layout for c in self.__controllers[self._VOTE]]
        l_stake = [c.layout.get_layout for c in self.__controllers[self._STAKE]]
        l_proposal = [c.layout.get_layout for c in self.__controllers[self._PROPOSAL]]
        l_organization = [c.layout.get_layout for c in self.__controllers[self._ORGANIZATION]]
        l_assets = [c.layout.get_layout for c in self.__controllers[self._ASSETS]]

        return {
            TEXT['activity_title'] : {
                'callables': l_organization,
                'css_id': TEXT['css_id_activity'],
                'disclaimer': TEXT['disclaimer_activity'],
            },
            TEXT['rep_holder_title']: {
                'callables': l_rep_h,
                'css_id': TEXT['css_id_reputation_holders'],
            },
            TEXT['proposal_title']: {
                'callables': l_proposal,
                'css_id': TEXT['css_id_proposal'],
            },
            TEXT['vote_title']: {
                'callables': l_vote,
                'css_id': TEXT['css_id_votes'],
            },
            TEXT['stake_title']: {
                'callables': l_stake,
                'css_id': TEXT['css_id_stake'],
            },
            TEXT['assets_title']: {
                'callables': l_assets,
                'css_id': TEXT['css_id_assets'],
            },
        }

    def __get_datapoints(self):
        if not self.are_panes:
            self.__gen_sections()

        return self.__data_points


    def __get_organization_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # active organizations
        charts.append(self.__create_chart(
            title=TEXT['title_active_organization'],
            adapter=MetricAdapter(s_factory.ACTIVE_ORGANIZATION),
            figure=CalFigure(),
            cont_key=self._ORGANIZATION,
            css_classes=['only-on-all-orgs'],
        ))

        charts.append(self.__create_chart(
            title=TEXT['title_organization_activity'],
            adapter=MetricAdapter(s_factory.ORGANIZATION_ACTIVITY),
            figure=CalFigure(),
            cont_key=self._ORGANIZATION
        ))

        return charts


    def __get_rep_holder_charts(self) -> List[Callable[[], html.Div]]:
        """
        Creates charts of reputation holder section, this includes 
         its layout and its controller.
        """
        charts: List[Callable] = list()

        # new reputation holders
        charts.append(self.__create_chart(
            title=TEXT['new_users_title'],
            adapter=MetricAdapter(s_factory.NEW_USERS),
            figure=BarFigure(),
            cont_key=self._REP_H
        ))

        # total reputation holders
        charts.append(self.__create_sum_chart(
            title=TEXT['total_users_title'],
            adapter=MetricAdapter(s_factory.TOTAL_REP_HOLDERS),
            figure=BarFigure(),
            cont_key=self._REP_H,
            dp_id=TEXT['total_users_dp_id'],
            dp_title=TEXT['total_users_dp_title'],
        ))

        # active reputation holders
        charts.append(self.__create_sum_chart(
            title=TEXT['active_users_title'],
            adapter=MetricAdapter(s_factory.ACTIVE_USERS),
            figure=BarFigure(),
            cont_key=self._REP_H,
            dp_id=TEXT['active_users_dp_id'],
            dp_title=TEXT['active_users_dp_title'],
        ))
        return charts


    def __get_vote_charts(self) -> List[Callable[[], html.Div]]:
        """
        Creates charts of vote section.
        """
        charts: List[Callable] = list()

        # total votes by type
        charts.append(self.__create_chart(
            title=TEXT['total_votes_option_title'],
            adapter=VoteType(s_factory.TOTAL_VOTES_OPTION, VoteType.VOTE),
            figure=MultiBarFigure(bar_type=MultiBarFigure.STACK),
            cont_key=self._VOTE
        ))

        # votes for rate
        charts.append(self.__create_chart(
            title=TEXT['vote_for_rate_title'],
            adapter=MetricAdapter(s_factory.VOTES_FOR_RATE),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()

        # votes against rate
        charts.append(self.__create_chart(
            title=TEXT['vote_against_rate_title'],
            adapter=MetricAdapter(s_factory.VOTES_AGAINST_RATE),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()

        # different voters
        charts.append(self.__create_chart(
            title=TEXT['different_voters_title'],
            adapter=MetricAdapter(s_factory.DIFFERENT_VOTERS),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))

        # percentage of reputation holders which vote
        charts.append(self.__create_chart(
            title=TEXT['voters_percentage_title'],
            adapter=MetricAdapter(s_factory.VOTERS_PERCENTAGE),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()
        self.__controllers[self._VOTE][-1].layout.figure\
            .configuration.add_y_params(params={
                'suffix': '%'})

        # vote-voters rate
        charts.append(self.__create_chart(
            title=TEXT['vote_voters_title'],
            adapter=MetricAdapter(s_factory.VOTE_VOTERS_RATE),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))

        return charts


    def __get_stake_charts(self) -> List[Callable[[], html.Div]]:
        """
        Creates charts of stake section.
        """
        charts: List[Callable] = list()

        # total stakes
        charts.append(self.__create_chart(
            title=TEXT['total_stakes_title'],
            adapter=MetricAdapter(s_factory.TOTAL_STAKES),
            figure=BarFigure(),
            cont_key=self._STAKE
        ))
        # different stakers
        charts.append(self.__create_chart(
            title=TEXT['different_stakers_title'],
            adapter=MetricAdapter(s_factory.DIFFERENT_STAKERS),
            figure=BarFigure(),
            cont_key=self._STAKE
        ))
        return charts


    def __get_proposal_charts(self) -> List[Callable[[], html.Div]]:
        """
        Creates charts of proposal section.
        """
        charts: List[Callable] = list()

        # new proposals
        charts.append(self.__create_sum_chart(
            title=TEXT['new_proposals_title'],
            adapter=MetricAdapter(s_factory.NEW_PROPOSALS),
            figure=BarFigure(),
            cont_key=self._PROPOSAL,
            dp_id=TEXT['new_proposals_dp_id'],
            dp_title=TEXT['new_proposals_dp_title'],
        ))

        # majority type
        charts.append(self.__create_chart(
            title=TEXT['proposal_outcome_majority_title'],
            adapter=MajorityType(s_factory.PROPOSAL_MAJORITY),
            figure=DoubleScatterFigure(),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()

        # proposal boost_outcome
        charts.append(self.__create_chart(
            title=TEXT['proposal_boost_outcome_title'],
            adapter=ProposalBoostOutcome(s_factory.PROPOSALS_BOOST_OUTCOME),
            figure=MultiBarFigure(bar_type=MultiBarFigure.STACK),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()

        # proposal approve rate
        charts.append(self.__create_chart(
            title=TEXT['approval_proposal_rate_title'],
            adapter=MetricAdapter(s_factory.APPROVAL_PROPOSAL_RATE),
            figure=BarFigure(),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()

        # total succes rate of the stakes
        charts.append(self.__create_chart(
            title=TEXT['proposal_total_succ_rate_title'],
            adapter=MetricAdapter(s_factory.PROPOSALS_TOTAL_SUCCES_RATIO),
            figure=BarFigure(),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()

        # success rate by type
        charts.append(self.__create_chart(
            title=TEXT['proposal_boost_succ_rate_title'],
            adapter=SuccessRatioType(s_factory.PROPOSALS_BOOST_SUCCES_RATIO),
            figure=MultiBarFigure(bar_type=MultiBarFigure.GROUP),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()

        return charts

    def __get_assets_charts(self):
        charts: List[Callable] = list()

        charts.append(self.__create_sum_chart(
            title=TEXT['assets_value_title'],
            adapter=AssetsValues(),
            figure=TreemapFigure(),
            cont_key=self._ASSETS,
            dp_id=TEXT['assets_value_dp_id'],
            dp_title=TEXT['assets_value_dp_title'],
        ))
        self.__controllers[self._ASSETS][-1].layout.configuration.disable_subtitles()

        charts.append(self.__create_dataTable(
            title=TEXT['assets_novalue_title'],
            adapter=AssetsTokens(),
            cont_key=self._ASSETS
        ))

        return charts

    def __create_chart(self, title: str, adapter: MetricAdapter, figure: Figure
    , cont_key: int, css_classes = []) -> Callable:
        """
        Creates the chart layout and its controller, and returns a callable
        to get the html representation.
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"
        layout: ChartPaneLayout = ChartPaneLayout(
            title=title,
            css_id=css_id,
            figure=figure,
            css_classes=css_classes,
        )
        layout.configuration.set_css_border(css_border=TEXT['css_pane_border'])

        controller: ChartController = ChartController(
            css_id=css_id,
            layout=layout,
            adapter=adapter)

        self.__controllers[cont_key].append(controller)
        return layout.get_layout

    def __create_sum_chart(self,
        title: str, 
        adapter: MetricAdapter, 
        figure: Figure, 
        cont_key: int,
        dp_id: str,
        dp_title: str,
    ) -> Callable:
        """
        Creates the chart layout and its controller, and returns a callable
        to get the html representation.
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"
        layout: ChartPaneLayout = ChartPaneLayout(
            title=title,
            css_id=css_id,
            figure=figure
        )
        layout.configuration.set_css_border(css_border=TEXT['css_pane_border'])

        self.__data_points[dp_id] = DataPointLayout(
            css_id=dp_id,
            title=dp_title,
        )

        controller: ChartController = ChartSummaryController(
            css_id=css_id,
            layout=layout,
            adapter=adapter,
            datapoint_layout=self.__data_points[dp_id]
        )

        self.__controllers[cont_key].append(controller)
        return layout.get_layout
    
    def __create_dataTable(self, title: str, adapter: IMetricAdapter, cont_key: int) -> Callable:
        """Creates a datatable to put alongside charts
        Args:
            title (str): The title of the datatable
            adapter (IMetricAdapter): The adapter to get the data from
            cont_key (int): The key of the controller
        Returns:
            Callable: Layout html builder
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"

        layout: DataTableLayout = DataTableLayout(
            title=title,
            css_id=css_id
        )

        controller: DataTableController = DataTableController(
            table_id=layout.table_id,
            layout=layout,
            adapter=adapter
        )

        self.__controllers[cont_key].append(controller)
        return layout.get_layout



"""
   Descp: This class is used to adapt the proposal_boost_outcome
          metric to its visual representation.

   Created on: 14-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict
from dao_analyzer.apps.common.business.transfers.organization.organization_list import OrganizationList

import dao_analyzer.apps.common.resources.colors as Color
from dao_analyzer.apps.daostack.resources.strings import TEXT
from dao_analyzer.apps.daostack.business.metric_adapter.metric_adapter import MetricAdapter
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
import dao_analyzer.apps.daostack.data_access.daos.metric.\
    metric_dao_factory as s_factory

class ProposalBoostOutcome(MetricAdapter):

    def __init__(self, metric_id: int) -> None:
        super().__init__(metric_id)

    
    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param.
        """
        dao = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=super().metric_id
        )
        metric: StackedSerie = dao.get_metric()

        y1 = metric.get_i_stack(0)
        y2 = metric.get_i_stack(1)
        y3 = metric.get_i_stack(2)
        y4 = metric.get_i_stack(3)

        return {
            'type1': {
                'y': y1,
                'color': Color.DARK_GREEN,
                'name': TEXT['queue_pass'],
            },
            'type2': {
                'y': y2,
                'color': Color.LIGHT_GREEN,
                'name': TEXT['boost_pass'],
            },
            'type3': {
                'y': y3,
                'color': Color.LIGHT_RED,
                'name': TEXT['boost_fail'],
            },
            'type4': {
                'y': y4,
                'color': Color.DARK_RED,
                'name': TEXT['queue_fail'],
            },
            'common': {
                'x': metric.get_serie(),
                'type': 'date',
                'x_format': MetricAdapter.DATE_FORMAT,
                'ordered_keys': ['type1', 'type2', 'type3', 'type4'],
            },
        }



"""
   Descp: Controller class for charts that also modify the summary info

   Created on: 29-apr-2022

   Copyright 2022 David Davó Laviña
        <ddavo@ucm.es>
"""
from dash.dependencies import Input, Output, State

from dao_analyzer.apps.common.business.i_metric_adapter import IMetricAdapter
from dao_analyzer.apps.common.business.transfers.organization import OrganizationList
from dao_analyzer.apps.common.presentation.data_point_layout import DataPointLayout
from dao_analyzer.apps.common.presentation.charts.layout.chart_pane_layout import ChartPaneLayout
from .chart_controller import ChartController

class ChartSummaryController(ChartController):
    def __init__(self, 
        css_id: str, 
        layout: ChartPaneLayout, 
        adapter: IMetricAdapter,
        datapoint_layout: DataPointLayout,
    ):
        super().__init__(css_id, layout, adapter)
        self._dp_layout = datapoint_layout

    @property
    def _dp_id(self) -> str:
        return self._dp_layout._id

    @property
    def _number_css_id(self) -> str:
        return f'{self._dp_id}-number'

    @property
    def _evolution_css_id(self) -> str:
        return f'{self._dp_id}-evolution'

    def bind_callback(self, app) -> None:

        @app.callback(
            Output(self._css_id, 'children'),
            Output(self._dp_id, 'children'),
            Input('org-dropdown', 'value'),
            State('org-dropdown', 'options'),
        )
        def update_chart(org_id, org_options):
            if not org_id:
                return self._layout.fill_child(), self._dp_layout.fill_child()
            
            data = self._adapter.get_plot_data(org_id, OrganizationList.from_dict_representation(org_options))

            number = None
            if 'last_value' in data:
                number = data['last_value']
            elif 'total' in data:
                number = data['total']

            evolution = data.get('diff', None)
            evolution_rel = data.get('diff_rel', None)

            return self._layout.fill_child(data), self._dp_layout.fill_child(number, evolution, evolution_rel)



import pandas as pd
from dash.dash_table.Format import Format, Scheme

from . import IMetricStrategy
from .....business.transfers import TabularData

class StAssetsTokens(IMetricStrategy):
    __GB_COLUMNS = ['network', 'symbol']
    __ALLOWED_COLUMNS = __GB_COLUMNS + ['balanceFloat']
    __VALUE_COLUMN = 'usdValue'

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        mask = ~(df[self.__VALUE_COLUMN] > 0)
        return df[mask][self.__ALLOWED_COLUMNS]

    def process_data(self, df: pd.DataFrame) -> TabularData:
        df = self.clean_df(df)
        df['cnt'] = 1

        dfg = df.groupby(self.__GB_COLUMNS)
        names = {
            'network': 'Network',
            'symbol': 'Symbol',
            'balanceFloat': 'Total Balance',
            'cnt': '# DAOs holding',
        }

        format
        formats = {
            'balanceFloat': Format(precision=5, scheme=Scheme.decimal_or_exponent)
        }

        sum = dfg.sum()

        # Remove the cnt if they are all from the same DAO
        if (sum['cnt'] == 1).all():
            sum = sum.drop(columns='cnt')
        
        td = TabularData.from_df(sum).set_column_names(names).set_column_formats(formats)
        return td


"""
    Descp: Daostack Runner and Collectors

    Created on: 15-nov-2021

    Copyright 2021 David Davó
        <david@ddavo.me>
"""
from typing import List

import pandas as pd
from gql.dsl import DSLField

from cache_scripts.common.blockscout import BlockscoutBallancesCollector
from cache_scripts.common.cryptocompare import CCPricesCollector

from ..common import ENDPOINTS, Collector
from ..common.graphql import GraphQLCollector, GraphQLRunner, add_where

def _changeProposalColumnNames(df: pd.DataFrame) -> pd.DataFrame:
    df = df.rename(columns={
        'daoId': 'dao',
        'proposalId': 'proposal'
    })
    return df

class BalancesCollector(BlockscoutBallancesCollector):
    def __init__(self, runner, base, network: str):
        super().__init__(runner, base=base, network=network, addr_key='dao')

class DaosCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('daos', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        
        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            df = df.rename(columns={
                'nativeTokenId':'nativeToken',
                'nativeReputationId':'nativeReputation'})
            return df
        
        @self.postprocessor
        def clone_id(df: pd.DataFrame) -> pd.DataFrame:
            df['dao'] = df['id']
            return df

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.daos(**add_where(kwargs, register="registered")).select(
            ds.DAO.id,
            ds.DAO.name,
            ds.DAO.nativeToken.select(ds.Token.id),
            ds.DAO.nativeReputation.select(ds.Rep.id)
        )

class ProposalsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('proposals', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])

        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            return df.rename(columns={
                'daoId': 'dao',
                'genesisProtocolParamsQueuedVoteRequiredPercentage': 'queuedVoteRequiredPercentage'
            })

        @self.postprocessor
        def deleteColums(df: pd.DataFrame) -> pd.DataFrame:
            return df.drop(columns=['competition'], errors='ignore')

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposals(**kwargs).select(
            ds.Proposal.id,
            ds.Proposal.proposer,
            # enum ProposalState { None, ExpiredInQueue, Executed, Queued, PreBoosted, Boosted, QuietEndingPeriod}
            ds.Proposal.stage,
            ds.Proposal.createdAt,
            ds.Proposal.preBoostedAt,
            ds.Proposal.boostedAt,
            ds.Proposal.closingAt,
            ds.Proposal.executedAt,
            ds.Proposal.totalRepWhenExecuted,
            ds.Proposal.totalRepWhenCreated,
            ds.Proposal.executionState,
            ds.Proposal.expiresInQueueAt,
            ds.Proposal.votesFor,
            ds.Proposal.votesAgainst,
            ds.Proposal.winningOutcome,
            ds.Proposal.stakesFor,
            ds.Proposal.stakesAgainst,
            ds.Proposal.genesisProtocolParams.select(ds.GenesisProtocolParam.queuedVoteRequiredPercentage),
            ds.Proposal.dao.select(ds.DAO.id),
            ds.Proposal.competition.select(ds.CompetitionProposal.id)
        )

class ReputationHoldersCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('reputationHolders', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.postprocessor(_changeProposalColumnNames)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.reputationHolders(**kwargs).select(
            ds.ReputationHolder.id,
            ds.ReputationHolder.contract,
            ds.ReputationHolder.address,
            ds.ReputationHolder.balance,
            ds.ReputationHolder.createdAt,
            ds.ReputationHolder.dao.select(ds.DAO.id)
        )

class StakesCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('stakes', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.postprocessor(_changeProposalColumnNames)
    
    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposalStakes(**kwargs).select(
            ds.ProposalStake.id,
            ds.ProposalStake.createdAt,
            ds.ProposalStake.staker,
            ds.ProposalStake.outcome,
            ds.ProposalStake.amount,
            ds.ProposalStake.dao.select(ds.DAO.id),
            ds.ProposalStake.proposal.select(ds.Proposal.id)
        )

class TokenPricesCollector(CCPricesCollector):
    pass

class VotesCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('votes', runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.postprocessor(_changeProposalColumnNames)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposalVotes(**kwargs).select(
            ds.ProposalVote.id,
            ds.ProposalVote.createdAt,
            ds.ProposalVote.voter,
            ds.ProposalVote.outcome,
            ds.ProposalVote.reputation,
            ds.ProposalVote.dao.select(ds.DAO.id),
            ds.ProposalVote.proposal.select(ds.Proposal.id)
        )

class CommonRepEventCollector(GraphQLCollector):
    def __init__(self, name, runner, base, network: str): 
        super().__init__(name, runner, network=network, endpoint=ENDPOINTS[network]['daostack'])
        self.base = base

        @self.postprocessor
        def add_dao_id(df: pd.DataFrame) -> pd.DataFrame:
            """ Using the contract info, appends the DAO id.
            Used by ReputationMintsCollector and ReputationBurnsCollector
            """
            # Skip postprocessor if empty
            if df.empty:
                return df
            
            l_index = ['network', 'contract']
            r_index = ['network', 'nativeReputation']

            wants = ['dao']
            prev_cols = list(df.columns)

            # Add the DAO field to the dataframe
            df = df.merge(self.base.df[r_index + wants],
                how='left',
                left_on=l_index,
                right_on=r_index,
            )

            # Get only the dao field
            df = df[prev_cols + wants]
            
            return df

class ReputationMintsCollector(CommonRepEventCollector):
    def __init__(self, *args, **kwargs):
        super().__init__('reputationMints', *args, **kwargs)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.reputationMints(**add_where(kwargs, amount_not=0)).select(
            ds.ReputationMint.id,
            # ds.ReputationMint.txHash, # Not used
            ds.ReputationMint.contract,
            ds.ReputationMint.address,
            ds.ReputationMint.amount,
            ds.ReputationMint.createdAt
        )

class ReputationBurnsCollector(CommonRepEventCollector):
    def __init__(self, *args, **kwargs):
        super().__init__('reputationBurns', *args, **kwargs)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.reputationBurns(**add_where(kwargs, amount_not=0)).select(
            ds.ReputationBurn.id,
            # ds.ReputationBurn.txHash, # Not used
            ds.ReputationBurn.contract,
            ds.ReputationBurn.address,
            ds.ReputationBurn.amount,
            ds.ReputationBurn.createdAt
        )

class DaostackRunner(GraphQLRunner):
    name: str = 'daostack'

    def __init__(self):
        super().__init__()
        self._collectors: List[Collector] = []
        for n in self.networks:
            dc = DaosCollector(self, n)

            self._collectors.extend([
                dc,
                ProposalsCollector(self, n),
                ReputationHoldersCollector(self, n),
                StakesCollector(self, n),
                VotesCollector(self, n),
                BalancesCollector(self, dc, n),
                ReputationMintsCollector(self, dc, n),
                ReputationBurnsCollector(self, dc, n),
            ])

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors



"""
   Descp: Strategy pattern to create a votation outcome metric.

   Created on: 21-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.serie import Serie
import dao_analyzer.apps.common.data_access.pandas_utils as pd_utl


class StVoteOutcome(IMetricStrategy):
    __DECIMALS = 10000000000000000
    __DF_DATE = 'startDate'
    __DF_EXECUTED = 'executed'
    __DF_YEA = 'yea'
    __DF_NAY = 'nay'
    __DF_SUPPORT = 'supportRequiredPct'
    __DF_VOTING = 'votingPower'
    __DF_QUORUM = 'minAcceptQuorum'
    __DF_PASS = 'didPass'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_YEA, __DF_NAY, __DF_SUPPORT, __DF_VOTING, __DF_QUORUM]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        # just take closed proposals
        # dff = pd_utl.filter_by_col_value(
        #     df=dff, 
        #     col=self.__DF_EXECUTED, 
        #     value=True,
        #     filters=[pd_utl.EQ])

        dff.loc[:, self.__DF_COLS] = dff[self.__DF_COLS]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)
        df = self.__calculate_outcome(df=df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        df = pd_utl.count_cols_repetitions(df, [self.__DF_DATE, self.__DF_PASS], self.__DF_COUNT)
        
        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        serie: Serie = Serie(x=dff[self.__DF_DATE].tolist())
        passes: List = self.__get_outcome(True, df, dff)
        rejections: List = self.__get_outcome(False, df, dff)

        metric: StackedSerie = StackedSerie(
            serie = serie,
            y_stack = [rejections, passes])

        return metric


    def __get_outcome(self, has_pass: bool, df: pd.DataFrame, 
    dff: pd.DataFrame) -> List[int]:

        d3f: pd.DataFrame = pd_utl.filter_by_col_value(
            df=df, 
            col=self.__DF_PASS, 
            value=has_pass,
            filters=[pd_utl.EQ])

        d3f = d3f.drop(columns=[self.__DF_PASS])
        d3f = pd.concat([d3f, dff], ignore_index=True)
        d3f = d3f.drop_duplicates(subset=self.__DF_DATE, keep="first")
        d3f = d3f.sort_values(self.__DF_DATE)

        return d3f[self.__DF_COUNT].tolist()


    def __calculate_outcome(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        outcome = ((positive voted tokens) / total voted tokens * 100) >= support required
                and
                ((positive voted tokens) / total tokens * 100) >= minimum accepted quorum
        """
        dff: pd.DataFrame = df

        has_pass: List[bool] = []
        yea: List[int] = [int(x) for x in dff[self.__DF_YEA].tolist()]
        nay: List[int] = [int(x) for x in dff[self.__DF_NAY].tolist()]
        total: List[int] = [int(x) for x in dff[self.__DF_VOTING].tolist()]
        l_support: List[int] = [int(x) for x in dff[self.__DF_SUPPORT].tolist()]
        l_quorum: List[int] = [int(x) for x in dff[self.__DF_QUORUM].tolist()]

        for i in range(len(dff)):
            support: int = self.__percentage(l_support[i])
            quorum: int = self.__percentage(l_quorum[i])

            positive_support: int = self.__ratio(yea[i], (yea[i] + nay[i])) * 100
            total_support: int = self.__ratio(yea[i], total[i]) * 100
            
            has_pass.append( (positive_support >= support) and (total_support >= quorum) )

        dff.loc[:, self.__DF_PASS] = has_pass
        dff.loc[:, [self.__DF_DATE, self.__DF_PASS]] = \
            dff[[self.__DF_DATE, self.__DF_PASS]]

        return dff


    def __percentage(self, value: int) -> int:
        """
        transform big int Aragon percentage into a normal percentage
        """
        return value / self.__DECIMALS


    def __ratio(self, num: int, divider: int) -> int:
        if divider == 0:
            return 0

        return num / divider



"""
   Descp: Vote outcome metric strategy test.

   Created on: 21-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.\
    st_vote_outcome import StVoteOutcome

from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie


class StVoteOutcomeTest(unittest.TestCase):

    def __check_lists(self, df: pd.DataFrame, out: List[List[int]]) -> None:
        strategy: StVoteOutcome = StVoteOutcome()
        result: StackedSerie = strategy.process_data(df=df)

        self.assertListEqual(out[0], result.get_i_stack(i_stack=0)) # test rejections
        self.assertListEqual(out[1], result.get_i_stack(i_stack=1)) # test passes


    def test_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=16, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'yea': 0, 'nay': 1, 'supportRequiredPct': 15, 'minAcceptQuorum': 100000000000000000, 'votingPower': 5,'startDate': bl.unix()},#today_year-(today_month-3)-16T00:00:00+00:00
            {'yea': 6, 'nay': 2, 'supportRequiredPct': 50, 'minAcceptQuorum': 50, 'votingPower': 9,'startDate': bl.add(month=1).change(day=5).unix()},#today_year-(today_month-2)-05T00:00:00+00:00
            {'yea': 1, 'nay': 0, 'supportRequiredPct': 100000000000000000, 'minAcceptQuorum': 330000000000000000, 'votingPower': 3,'startDate': bl.unix()},#today_year-(today_month-2)-05T00:00:00+00:00
            {'yea': 5, 'nay': 4, 'supportRequiredPct': 330000000000000000, 'minAcceptQuorum': 150000000000000000, 'votingPower': 9,'startDate': bl.add(month=2).change(day=1, hour=23, minute=59, second=59).unix()},#today_year-today_month-01T23:59:59+00:00
            {'yea': 0, 'nay': 0, 'supportRequiredPct': 0, 'minAcceptQuorum': 1, 'votingPower': 00000000000000000,'startDate': bl.unix()},#today_year-today_month-01T23:59:59+00:00
            {'yea': 4, 'nay': 0, 'supportRequiredPct': 1000000000000000000, 'minAcceptQuorum': 500000000000000000, 'votingPower': 10,'startDate': bl.add(second=1).change(day=27).unix()},#today_year-today_month-27T00:00:00+00:00
        ])
        out: List[List[int]] = [
            [1, 0, 0, 2], # number of rejected proposals
            [0, 2, 0, 1] # number of approved proposals
        ]

        self.__check_lists(df=in_df, out=out)


if __name__ == "__main__":
    unittest.main()



from dash import html

from dao_analyzer.apps.common.resources.strings import TEXT

class DataPointLayout:
    def __init__(self, css_id, title):
        self._id = css_id
        self._title = title

    @staticmethod
    def _get_dp_icon(evolution: float):
        if evolution is None:
            return

        try:
            if evolution < 0:
                return html.I(className="bi bi-arrow-down-circle-fill dp-icon-down")
            elif evolution > 0:
                return html.I(className="bi bi-arrow-up-circle-fill dp-icon-up")
            else:
                return html.I(className="bi bi-dash-circle dp-icon-same")
        except ValueError:
            return

    def fill_child(
        self,
        number: str = "?",
        evolution: float = None,
        evolution_rel: float = None,
    ):
        evolution_children = []
        if evolution is not None or evolution_rel is not None:
            evolution_children.append(self._get_dp_icon(evolution or evolution_rel))
            evolution_children.append(' ')

        if evolution is not None and evolution_rel is not None:
            evolution_children.append(f'{evolution:.0f} ({abs(evolution_rel):.2f}%)')
        elif evolution is not None:
            evolution_children.append(f'{evolution:.0f}')
        elif evolution_rel is not None:
            evolution_children.append(f'{evolution_rel:.2f}')

        return [
            html.Span(self._title, className="dao-summary-datapoint-title"),
            html.Div(
                number, className="dao-summary-datapoint-number", id=self._id + "-number"
            ),
            html.Div(
                TEXT["this_month"],
                className="dao-summary-datapoint-lastmonth",
            ) if evolution_children else None,
            html.Div(
                evolution_children,
                className="dao-summary-datapoint-evolution",
                id=self._id + "-evolution",
            ) if evolution_children else None,
        ]

    def get_layout(self) -> html.Div:
        return html.Div(self.fill_child(), className='dao-summary-datapoint', id=self._id)




"""
   Descp: Wraps the figure and the chart pane configuration.

   Created on: 14-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List

from dao_analyzer.apps.common.presentation.charts.layout.figure.figure_configuration \
    import FigureConfiguration
from .layout_configuration import LayoutConfiguration

class ChartPaneConfiguration(LayoutConfiguration, FigureConfiguration):

    def __init__(self, fig_conf: FigureConfiguration) -> None:
        super().__init__()
        self.__fig_conf = fig_conf
        self.__show_subtitles: bool = True

    @property
    def fig_conf(self) -> FigureConfiguration:
        return self.__fig_conf


    @property
    def show_subtitles(self) -> bool:
        return self.__show_subtitles

    def enable_legend(self) -> None:
        self.fig_conf.enable_legend()


    def disable_legend(self) -> None:
        self.fig_conf.disable_legend()

    
    def get_legend(self) -> Dict:
        return self.fig_conf.get_legend()


    def add_horizontal_line(self, y: float, y_ref: str) -> None:
        self.fig_conf.add_horizontal_line(y, y_ref)


    def get_shapes(self) -> List[Dict]:
        return self.fig_conf.get_shapes()


    def get_axis_layout(self, args: Dict) -> Dict:
        return self.fig_conf.get_axis_layout(args)


    def enable_subtitles(self) -> None:
        self.__show_subtitles = True


    def disable_subtitles(self) -> None:
        self.__show_subtitles = False



"""
    Descp: Daohaus Runner and Collectors

    Created on: 13-nov-2021

    Copyright 2021 David Davó
        <david@ddavo.me>
"""
import requests
import requests_cache
from typing import List
from datetime import timedelta

import pandas as pd
from tqdm import tqdm
from gql.dsl import DSLField

from .. import config
from ..common.common import solve_decimals
from ..common.cryptocompare import cc_postprocessor

from ..common import ENDPOINTS, Collector
from ..common.graphql import GraphQLCollector, GraphQLRunner, add_where

DATA_ENDPOINT: str = "https://data.daohaus.club/dao/{id}"

class MembersCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('members', runner, network=network, endpoint=ENDPOINTS[network]['daohaus'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.members(**kwargs).select(
            ds.Member.id,
            ds.Member.createdAt,
            ds.Member.molochAddress,
            ds.Member.memberAddress,
            ds.Member.shares,
            ds.Member.loot,
            ds.Member.exists,
            ds.Member.didRagequit
        )

class MolochesCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('moloches', runner, network=network, endpoint=ENDPOINTS[network]['daohaus_stats'])

        @self.postprocessor
        def moloch_id(df: pd.DataFrame) -> pd.DataFrame:
            df['molochAddress'] = df['id']
            return df

        @self.postprocessor
        def moloch_names(df: pd.DataFrame) -> pd.DataFrame:
            df = df.rename(columns={"title":"name"})

            if config.skip_daohaus_names:
                return df

            cached = requests_cache.CachedSession(self.data_path.parent / '.names_cache', 
                use_cache_dir=False, 
                expire_after=timedelta(days=30)
            )
            tqdm.pandas(desc="Getting moloch names")
            df["name"] = df.progress_apply(lambda x:self._request_moloch_name(cached, x['id']), axis=1)

            return df
    
    @staticmethod
    def _request_moloch_name(req: requests.Session, moloch_id: str):
        response = req.get(DATA_ENDPOINT.format(id=moloch_id))

        o = response.json()
        if isinstance(o, list) and o and "name" in o[0]:
            return o[0]["name"]
        else:
            return None
    
    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.moloches(**add_where(kwargs, deleted=False)).select(
            ds.Moloch.id,
            ds.Moloch.title,
            ds.Moloch.version,
            ds.Moloch.summoner,
            ds.Moloch.summoningTime,
            ds.Moloch.timestamp,
            ds.Moloch.proposalCount,
            ds.Moloch.memberCount,
            ds.Moloch.voteCount,
            ds.Moloch.rageQuitCount,
            ds.Moloch.totalGas
        )

class ProposalsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('proposals', runner, network=network, endpoint=ENDPOINTS[network]["daohaus"])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.proposals(**kwargs).select(
            ds.Proposal.id,
            ds.Proposal.createdAt,
            ds.Proposal.proposalId,
            ds.Proposal.molochAddress,
            ds.Proposal.memberAddress,
            ds.Proposal.proposer,
            ds.Proposal.sponsor,
            ds.Proposal.sharesRequested,
            ds.Proposal.lootRequested,
            ds.Proposal.tributeOffered,
            ds.Proposal.paymentRequested,
            ds.Proposal.yesVotes,
            ds.Proposal.noVotes,
            ds.Proposal.sponsored,
            ds.Proposal.sponsoredAt,
            ds.Proposal.processed,
            ds.Proposal.processedAt,
            ds.Proposal.didPass,
            ds.Proposal.yesShares,
            ds.Proposal.noShares
        )
    
class RageQuitCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('rageQuits', runner, network=network, endpoint=ENDPOINTS[network]["daohaus"])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.rageQuits(**kwargs).select(
            ds.RageQuit.id,
            ds.RageQuit.createdAt,
            ds.RageQuit.molochAddress,
            ds.RageQuit.memberAddress,
            ds.RageQuit.shares,
            ds.RageQuit.loot
        )

class TokenBalancesCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('tokenBalances', runner, network=network, endpoint=ENDPOINTS[network]["daohaus"])

        @self.postprocessor
        def change_col_names(df: pd.DataFrame) -> pd.DataFrame:
            return df.rename(columns={
                'molochId': 'molochAddress',
                'tokenTokenAddress': 'tokenAddress',
                'tokenDecimals': 'decimals',
                'tokenBalance': 'balance',
                'tokenSymbol': 'symbol'
            })

        @self.postprocessor
        def coalesce_bank_type(df: pd.DataFrame) -> pd.DataFrame:
            bank_idx = ['guildBank', 'memberBank', 'ecrowBank']

            df['bank'] = df[bank_idx].idxmax(1)
            df['bank'] = df['bank'].str.lower()
            df['bank'] = df['bank'].str.replace('bank', '')
            df = df.drop(columns=bank_idx)
        
            return df
        
        self.postprocessors.append(solve_decimals)
        self.postprocessors.append(cc_postprocessor)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.tokenBalances(**add_where(kwargs, guildBank=True, tokenBalance_gt=0)).select(
            ds.TokenBalance.id,
            ds.TokenBalance.moloch.select(
                ds.Moloch.id
            ),
            ds.TokenBalance.token.select(
                ds.Token.tokenAddress,
                ds.Token.symbol,
                ds.Token.decimals
            ),
            ds.TokenBalance.guildBank,
            ds.TokenBalance.memberBank,
            ds.TokenBalance.ecrowBank,
            ds.TokenBalance.tokenBalance
        )

class VoteCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('votes', runner, network=network, endpoint=ENDPOINTS[network]["daohaus"])

        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            return df.rename(columns={"proposalId":"proposalAddress"})

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.votes(**kwargs).select(
            ds.Vote.id,
            ds.Vote.createdAt,
            ds.Vote.proposal.select(ds.Proposal.id),
            ds.Vote.molochAddress,
            ds.Vote.memberAddress,
            ds.Vote.uintVote
        )

class DaohausRunner(GraphQLRunner):
    name: str = 'daohaus'

    def __init__(self):
        super().__init__()
        self._collectors: List[Collector] = []
        for n in self.networks:
            self._collectors.extend([
                MembersCollector(self, n),
                MolochesCollector(self, n),
                ProposalsCollector(self, n),
                RageQuitCollector(self, n),
                TokenBalancesCollector(self, n),
                VoteCollector(self, n)
            ])

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors


from typing import List
import pandas as pd
from pathlib import Path
import portalocker as pl
from datetime import datetime, timedelta, timezone
import json
import logging

from dao_analyzer.apps.common.business.singleton import ABCSingleton
from dao_analyzer.apps.common.data_access.requesters.irequester import IRequester

LOCK_PATH = Path('datawarehouse/.lock')

class CacheRequester(IRequester, metaclass=ABCSingleton):
    CHECKING_COOLDOWN = 60

    def __init__(self, srcs: List[Path]):
        """
        Initializes the CacheRequester

            Parameters:
                srcs (List[Path]): A list of paths to read the dataframes from
        """
        self._srcs = srcs
        self._df: pd.DataFrame = pd.DataFrame()

        # Avoids having to lock the file and open the metadata with every request
        self._next_check = datetime.min

        # Lets us read the data only when necessary
        self._last_update = datetime.min.replace(tzinfo=timezone.utc)
        self.logger = logging.getLogger("app.cacherequester")

    def get_last_update(self) -> datetime:
        return self._last_update
    
    def get_last_update_str(self) -> str:
        return self.get_last_update().strftime('%b %-d, %Y at %H:%M %Z')

    def metadataTime(self) -> datetime:
        """
        Returns the latest time the dataframes have been updated
        """
        t = datetime.min.replace(tzinfo=timezone.utc)
        metadata = json.loads((self._srcs[0].parent / 'metadata.json').read_bytes())["metadata"]

        for src in self._srcs:
            collector_name = str(src.stem)
            for k,v in metadata.items():
                # Format is <platform>/<collector>-<network>
                if k.split('/')[1].split('-')[0] == collector_name:
                    t = max(t, datetime.fromisoformat(v["block"]["timestamp"]))

        return t

    def tryReload(self):
        self._df = pd.concat(map(pd.read_feather, self._srcs), axis=0, ignore_index=True)

    # https://stackoverflow.com/questions/70861731/how-to-filelock-an-entire-directory
    def request(self) -> pd.DataFrame:
        """
        Gets data from datawarehouse.
        Arguments:
            * args: Its not used
        Return:
            a pandas dataframe with all the data loaded. If the src does not 
            exist, it will return an empty dataframe.
        """
        # Whether we should check (locking can be costly)
        if self._srcs and (self._df.empty or datetime.now() > self._next_check):
            self._next_check = datetime.now() + timedelta(seconds=self.CHECKING_COOLDOWN)

            # Try checking if something has changed
            try:
                # Lock the datawarehouse as reader, and then check if the metadata changed
                with pl.Lock(LOCK_PATH, 'rb', flags=pl.LockFlags.SHARED  | pl.LockFlags.NON_BLOCKING, timeout=0.1):
                    t = self.metadataTime()
                    if t > self._last_update or self._df.empty:
                        self._last_update = t
                        self.tryReload()
            except (pl.LockException, pl.AlreadyLocked):
                self.logger.debug("Couldn't acquire lock")

        if self._df.empty:
            self.logger.warning(f"Returning empty dataframe for sources {[str(x) for x in self._srcs]}")

        return self._df



"""
   Descp: Serie of values transfer.

   Created on: 3-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List
from datetime import date

from dao_analyzer.apps.common.resources.strings import TEXT

class Serie():
    """
    * __x = a list of values which represent the serie.
    """

    def __init__(self, x: List = None):
        self.__x: List = x if x else list()

    
    def get_last_serie_elem(self) -> str:
        if not self.__x:
            return TEXT['no_data']

        val = f'{self.__x[-1]}'

        if type(self.__x[-1]) == date:
            val = self.__x[-1].strftime('%B')
                
        return val 


    def get_x(self) -> List:
        return self.__x



"""
   Descp: Main view controller to manage the callbacks.

   Created on: 1-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import List, Dict, Any

import dash
from dash import dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dao_analyzer.apps.common.business.transfers.organization.organization_filter import ALL_NETWORKS_VALUE
from dao_analyzer.apps.common.business.transfers.organization.organization_list import OrganizationList

from dao_analyzer.apps.common.presentation.main_view.main_view import generate_layout
import dao_analyzer.apps.common.presentation.about_view.about_view as about
import dao_analyzer.apps.daostack.business.app_service as daostack
import dao_analyzer.apps.daohaus.business.app_service as daohaus
import dao_analyzer.apps.aragon.business.app_service as aragon
from dao_analyzer.apps.common.resources.strings import TEXT

ABOUT_SUBPAGE: str = '__about'

def _process_params(search: str) -> Dict[str, str]:
    # TODO: Return Dict[str, Any] instead (parse json)
    search = search.removeprefix('?')

    if not search:
        return {}
    
    params = dict()
    for p in search.split('&'):
        k,v = p.split('=', 1) # Split only on first ocurrence
        params[k] = v

    return params

def _params_string(d: Dict[str, str]):
    if not d: 
        return ""
    
    return '?' + '&'.join(['='.join([k,v]) for k,v in d.items()])

def bind_callbacks(app) -> None: # noqa: C901

    # Callbacks need to be loaded twice.
    services: Dict[str, Any] = {
        "daostack": daostack.DaostackService(),
        "daohaus": daohaus.DaohausService(),
        "aragon": aragon.AragonService(),
    }

    for s in services.values():
        s.bind_callbacks(app)

    @app.callback(
        Output('page-content', 'children'),
        Output('header-loading-state', 'children'), # <- THIS CAUSES THE FLASH
        Input('page-content', 'data-subpage'),
        State('page-content', 'data-org-id'),
        State('url', 'search')
    )
    def change_subpage(subpage, org_id, search):
        if not subpage:
            return dcc.Location(pathname='/daohaus', id='default_redirect'), 'redirect'
        elif subpage == ABOUT_SUBPAGE:
            return generate_layout(body=about.get_layout()), ''
        elif subpage in services:
            return generate_layout(body=services[subpage].get_layout(org_value=org_id, network_value=_process_params(search).get('network', ''))), 'loading'

    @app.callback(
        Output('page-content', 'data-subpage'),
        Output('page-content', 'data-org-id'),
        Input('url', 'pathname'),
        State('page-content', 'data-subpage')
    )
    def url_changed(pathname, current_platform):
        if pathname == "/":
            return '', dash.no_update

        if pathname == '/' + TEXT['url_about']:
            return ABOUT_SUBPAGE, dash.no_update

        patharr = pathname.split("/")

        platform = patharr[1]
        org_id = patharr[2] if len(patharr) >= 3 else None

        # Dont regenerate layout if we are already using it
        if platform == current_platform:
            raise PreventUpdate

        return platform, org_id

    @app.callback(
        Output('url', 'pathname'),
        Input('daostack-bt', 'n_clicks'),
        Input('daohaus-bt', 'n_clicks'),
        Input('aragon-bt', 'n_clicks'),
        Input('org-dropdown', 'value'),
        State('url', 'pathname')
    )
    def load_ecosystem(bt_daostack: int, bt_daohaus: int, bt_aragon: int, dropdown_value: str, prev_pathname: str) -> str:
        ctx = dash.callback_context

        # prev_pathname state changed (maybe we did it)
        if not bt_daostack and not bt_daohaus and not bt_aragon and not dropdown_value:
            raise PreventUpdate
        # dropdown_value changed
        elif not bt_daostack and not bt_daohaus and not bt_aragon:
            platform = prev_pathname.split("/")[1]
            organizations = services[platform].organization_list()

            new_pathname = f'/{platform}'

            if not organizations.is_all_orgs(dropdown_value):
                new_pathname += f'/{dropdown_value}'

            if new_pathname == prev_pathname:
                raise PreventUpdate

            return new_pathname
        
        trigger = ctx.triggered_id
        
        pathname: str = "/"
        if trigger == 'daostack-bt':
            pathname = TEXT['url_daostack']
        elif trigger == 'daohaus-bt':
            pathname = TEXT['url_daohaus']
        elif trigger == 'aragon-bt':
            pathname = TEXT['url_aragon']

        return "/" + pathname

    @app.callback(
        Output('org-dropdown', 'options'),
        Output('org-dropdown', 'value'),
        Output('org-number', 'children'),
        Output('platform-info-store', 'data'),
        Output('url', 'search'),
        Input('org-filter', 'value'),
        Input('org-network-radio', 'value'),
        State('org-dropdown', 'value'),
        State('organization-list-store', 'data'),
        State('page-content', 'data-subpage'),
        State('platform-info-store', 'data'),
    )
    def org_filters(filter_values: List[str], network_value: str, org_value: str, org_list: list, platform_name: str, prev_platform: str):
        filtered = OrganizationList.from_json(org_list)

        # First we initialize all values
        organizations = filtered.filter(filter_values, network_value, only_enabled=True)
        options = organizations.get_dict_representation()
        org_number = f"There are {len(organizations):,} DAOs"
        platform = dash.no_update
        params = {}
        
        if network_value != ALL_NETWORKS_VALUE:
            params = {'network': network_value}

        # If the selected DAO was filtered out, fall back to All DAOs
        if org_value in [ x['value'] for x in options ]:
            value = org_value
        else:
            value = organizations.get_all_orgs_dict()['value']

        # Change only if is all orgs, or its not stablished yet
        if value == organizations.ALL_ORGS_ID or not prev_platform:
            platform = services[platform_name].platform(organizations)

        return options, value, org_number, platform, _params_string(params)



"""
   Descp: Tester for organization transfers.

   Created on: 13-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import List, Dict
import unittest
from hypothesis import given, example, settings, strategies as st
from random import Random

from dao_analyzer.apps.common.business.transfers.organization import Organization
from dao_analyzer.apps.common.business.transfers.organization import OrganizationList


OrganizationStrategy = st.builds(
    Organization,
    o_id=st.text(min_size=2),
    name=st.text())

ALL_ORGS_ID = OrganizationList.ALL_ORGS_ID

class OrganizationTest(unittest.TestCase):

    def test_get_dict_representation_1(self):
        org1: Organization = Organization(o_id='222', name='test2', network="n")
        org2: Organization = Organization(o_id='111', name='test1', network="n")
        orgs: OrganizationList = OrganizationList(orgs=[org1, org2])
        result: List[Dict[str, str]] = orgs.get_dict_representation()

        sol: List[Dict[str, str]] = [
            {'value': ALL_ORGS_ID, 'label': 'All DAOs'},
            {'value': '111', 'label': 'test1 (111... n)'},
            {'value': '222', 'label': 'test2 (222... n)'},
        ]

        for i, r in enumerate(result):
            self.assertEqual(sol[i]['value'], r['value'])
            self.assertEqual(sol[i]['label'], r['label'])


    @example(orgs=list())
    @given(orgs=st.lists(OrganizationStrategy))
    @settings(max_examples=30)
    def test_get_dict_representation_2(self, orgs):
        l_org: OrganizationList = OrganizationList(orgs=orgs)
        result: List[Dict[str, str]] = l_org.get_dict_representation()

        sol: int = len(orgs) + 1 if len(orgs) > 0 else 0

        self.assertEqual(sol, len(result))


    @given(orgs=st.lists(OrganizationStrategy, min_size=1))
    @settings(max_examples=30)
    def test_get_ids_from_id_1(self, orgs: List[Organization]):
        l_org: OrganizationList = OrganizationList(orgs=orgs)
        org: Organization = Random().choice(orgs)
        ids: List[str] = l_org.get_ids_from_id(org.get_id())

        self.assertEqual(1, len(ids))
        self.assertEqual(org.get_id(), ids[0])


    @example(orgs=list())
    @given(orgs=st.lists(OrganizationStrategy))
    @settings(max_examples=30)
    def test_get_ids_from_id_2(self, orgs: List[Organization]):
        l_org: OrganizationList = OrganizationList(orgs=orgs)
        ids: List[str] = l_org.get_ids_from_id(ALL_ORGS_ID)

        self.assertEqual(len(orgs), len(ids))


if __name__ == "__main__":
    unittest.main()



"""
   Descp: Manage the application logic, and it's used to interconect the
        data_access and presentation layers.

   Created on: 2-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List, Callable
from dash import html
from dao_analyzer.apps.common.business.transfers import Platform, OrganizationList

import dao_analyzer.apps.common.presentation.dashboard_view.dashboard_view as view
import dao_analyzer.apps.common.presentation.dashboard_view.controller as view_cont
from dao_analyzer.apps.aragon.data_access.daos.platform_dao import AragonDAO
from dao_analyzer.apps.common.data_access.daos.platform_dao import PlatformDAO
from dao_analyzer.apps.aragon.business.metric_adapter.asset_values import AssetsValues
from dao_analyzer.apps.aragon.business.metric_adapter.asset_tokens import AssetsTokens
from dao_analyzer.apps.common.presentation.charts.chart_controller import ChartController
from dao_analyzer.apps.common.presentation.charts.chart_sum_controller import ChartSummaryController
from dao_analyzer.apps.common.presentation.charts.dt_controller import DataTableController
from dao_analyzer.apps.common.presentation.data_point_layout import DataPointLayout
from dao_analyzer.apps.common.presentation.charts.layout import ChartPaneLayout, DataTableLayout
from dao_analyzer.apps.common.presentation.charts.layout.figure import Figure, BarFigure, CalFigure, MultiBarFigure, TreemapFigure
import dao_analyzer.apps.aragon.data_access.daos.metric.metric_dao_factory as s_factory
from dao_analyzer.apps.common.business.i_metric_adapter import IMetricAdapter
from dao_analyzer.apps.common.business.singleton import Singleton
from dao_analyzer.apps.aragon.business.metric_adapter.basic_adapter import BasicAdapter
from dao_analyzer.apps.aragon.business.metric_adapter.installed_apps import InstalledApps
from dao_analyzer.apps.aragon.business.metric_adapter.cast_type import CastType
from dao_analyzer.apps.aragon.business.metric_adapter.vote_outcome import VoteOutcome

from dao_analyzer.apps.aragon.resources.strings import TEXT

class AragonService(metaclass=Singleton):
    _TOKEN_HOLDER: int = 0
    _VOTE: int = 1
    _CAST: int = 2
    _TRANSACTION: int = 3
    _APP: int = 4
    _ORGANIZATION: int = 5
    _ASSETS: int = 6

    def __init__(self):
        # app state
        self.__orgsDAO: PlatformDAO = AragonDAO()
        self.__controllers: Dict[int, List[ChartController]] = {
            self._TOKEN_HOLDER: list(),
            self._VOTE: list(),
            self._CAST: list(),
            self._TRANSACTION: list(),
            self._APP: list(),
            self._ORGANIZATION: list(),
            self._ASSETS: list()
        }
        self.__already_bound: bool = False
        self.__data_points: Dict[str, DataPointLayout] = {}
    
    def bind_callbacks(self, app) -> None:
        if not self.__already_bound:
            self.__already_bound = True
            view_cont.bind_callbacks(
                app=app,
                section_id=TEXT['css_id_organization'],
            )
            self.__gen_sections()

            for contList in self.__controllers.values():
                for c in contList:
                    if hasattr(c, 'bind_callback'):
                        c.bind_callback(app)


    def platform(self, orglist: OrganizationList) -> Platform:
        return self.__orgsDAO.get_platform(orglist)

    def organization_list(self) -> OrganizationList:
        return self.__orgsDAO.get_organization_list()

    @property
    def are_panes(self) -> bool:
        """
        Checks if panes and their controllers are already created.
        """
        return any(self.__controllers.values())


    def get_layout(self, **kwargs) -> html.Div:
        """
        Returns the app's layout. 
        """
        if not self.__already_bound:
            self.bind_callbacks()
        
        return view.generate_layout(
            organization_list=self.organization_list(),
            sections=self.__get_sections(),
            ecosystem='aragon',
            update=self.__orgsDAO.get_last_update_str(),
            platform_id=TEXT['css_id_organization'],
            datapoints=self.__get_datapoints(),
            **kwargs
        )
    

    def __gen_sections(self) -> None:
        self.__get_token_holder_charts()
        self.__get_vote_charts()
        self.__get_cast_charts()
        self.__get_transaction_charts()
        self.__get_app_charts()
        self.__get_organization_charts()
        self.__get_assets_charts()

    def __get_sections(self) -> Dict[str, List[Callable]]:
        """
        Returns a dict with each section filled with a callable function which
         returns the chart layout
        """
        l_token_holders: List[Callable] = list()
        l_vote: List[Callable] = list()
        l_cast: List[Callable] = list()
        l_transaction: List[Callable] = list()
        l_app: List[Callable] = list()
        l_organization: List[Callable] = list()
        l_assets: List[Callable] = list()

        if not self.are_panes:
            self.__gen_sections()

        # Panes are already created.
        l_token_holders = [c.layout.get_layout for c in self.__controllers[self._TOKEN_HOLDER]]
        l_vote = [c.layout.get_layout for c in self.__controllers[self._VOTE]]
        l_cast = [c.layout.get_layout for c in self.__controllers[self._CAST]]
        l_transaction = [c.layout.get_layout for c in self.__controllers[self._TRANSACTION]]
        l_app = [c.layout.get_layout for c in self.__controllers[self._APP]]
        l_organization = [c.layout.get_layout for c in self.__controllers[self._ORGANIZATION]]
        l_assets = [c.layout.get_layout for c in self.__controllers[self._ASSETS]]

        return {
            TEXT['title_section_activity']: {
                'callables': l_organization,
                'css_id': TEXT['css_id_activity'],
                'disclaimer': TEXT['disclaimer_activity'],
            },
            TEXT['title_section_token_holders']: {
                'callables': l_token_holders,
                'css_id': TEXT['css_id_token_holder'],
            },
            TEXT['title_section_vote']: {
                'callables': l_vote,
                'css_id': TEXT['css_id_vote'],
            },
            TEXT['title_section_cast']: {
                'callables': l_cast,
                'css_id': TEXT['css_id_cast']
            },
            TEXT['title_section_transaction']: {
                'callables': l_transaction,
                'css_id': TEXT['css_id_transactions'],
            },
            TEXT['title_section_app']: {
                'callables': l_app,
                'css_id': TEXT['css_id_app'],
            },
            TEXT['title_assets']: {
                'callables': l_assets,
                'css_id': TEXT['css_id_assets'],
                'disclaimer': TEXT['disclaimer_assets'],
            },
        }

    def __get_datapoints(self):
        if not self.are_panes:
            self.__gen_sections()
        
        return self.__data_points

    def __get_organization_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # active organization
        charts.append(self.__create_chart(
            title=TEXT['title_active_organization'],
            adapter=BasicAdapter(
                metric_id=s_factory.ACTIVE_ORGANIZATION, 
            ),
            figure=CalFigure(),
            cont_key=self._ORGANIZATION,
            css_classes=['only-on-all-orgs'],
        ))

        charts.append(self.__create_chart(
            title=TEXT['title_organization_activity'],
            adapter=BasicAdapter(
                metric_id=s_factory.ORGANIZATION_ACTIVITY,
            ),
            figure=CalFigure(),
            cont_key=self._ORGANIZATION,
        ))

        return charts


    def __get_token_holder_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # total token holders
        charts.append(self.__create_sum_chart(
            title=TEXT['title_total_token_holders'],
            adapter=BasicAdapter(
                metric_id=s_factory.TOTAL_TOKEN_HOLDERS,
            ),
            figure=BarFigure(),
            cont_key=self._TOKEN_HOLDER,
            dp_id=TEXT['dp_id_total_token_holders'],
            dp_title=TEXT['dp_title_total_token_holders'],
        ))

        # active token holders
        charts.append(self.__create_sum_chart(
            title=TEXT['title_active_token_holders'],
            adapter=BasicAdapter(
                metric_id=s_factory.ACTIVE_TOKEN_HOLDERS, 
            ),
            figure=BarFigure(),
            cont_key=self._TOKEN_HOLDER,
            dp_id=TEXT['dp_id_active_token_holders'],
            dp_title=TEXT['dp_title_active_token_holders'],
        ))

        return charts


    def __get_vote_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # new votes
        charts.append(self.__create_sum_chart(
            title=TEXT['title_new_votes'],
            adapter=BasicAdapter(
                metric_id=s_factory.NEW_VOTES, 
            ),
            figure=BarFigure(),
            cont_key=self._VOTE,
            dp_id=TEXT['dp_id_new_votes'],
            dp_title=TEXT['dp_title_new_votes'],
        ))

        # vote's outcome
        charts.append(self.__create_chart(
            title=TEXT['title_vote_outcome'],
            adapter=VoteOutcome(),
            figure=MultiBarFigure(MultiBarFigure.STACK),
            cont_key=self._VOTE
        ))

        # approval vote rate
        charts.append(self.__create_chart(
            title=TEXT['title_rate_vote_rate'],
            adapter=BasicAdapter(
                metric_id=s_factory.APPROVAL_VOTE_RATE, 
            ),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()

        return charts


    def __get_cast_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # cast type
        charts.append(self.__create_chart(
            title=TEXT['title_cast_type'],
            adapter=CastType(),
            figure=MultiBarFigure(bar_type=MultiBarFigure.STACK),
            cont_key=self._CAST
        ))

        # casted votes for rate
        charts.append(self.__create_chart(
            title=TEXT['title_casted_votes_for_rate'],
            adapter=BasicAdapter(
                metric_id=s_factory.CASTED_VOTE_FOR_RATE,
            ),
            figure=BarFigure(),
            cont_key=self._CAST
        ))
        self.__controllers[self._CAST][-1].layout.configuration.disable_subtitles()

        # casted votes against rate
        charts.append(self.__create_chart(
            title=TEXT['title_casted_votes_against_rate'],
            adapter=BasicAdapter(
                metric_id=s_factory.CASTED_VOTE_AGAINST_RATE,
            ),
            figure=BarFigure(),
            cont_key=self._CAST
        ))
        self.__controllers[self._CAST][-1].layout.configuration.disable_subtitles()

        # active voters
        charts.append(self.__create_chart(
            title=TEXT['title_active_voters'],
            adapter=BasicAdapter(
                metric_id=s_factory.ACTIVE_VOTERS,
            ),
            figure=BarFigure(),
            cont_key=self._CAST
        ))

        # casted votes-voters rate
        charts.append(self.__create_chart(
            title=TEXT['title_votes_voters'],
            adapter=BasicAdapter(
                metric_id=s_factory.CASTED_VOTE_VOTER_RATE,
            ),
            figure=BarFigure(),
            cont_key=self._CAST
        ))

        return charts


    def __get_transaction_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # new transactions
        charts.append(self.__create_chart(
            title=TEXT['title_new_transactions'],
            adapter=BasicAdapter(
                metric_id=s_factory.NEW_TRANSACTIONS, 
            ),
            figure=BarFigure(),
            cont_key=self._TRANSACTION
        ))
        return charts


    def __get_app_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # installed apps
        charts.append(self.__create_chart(
            title=TEXT['title_installed_apps'],
            adapter=InstalledApps(),
            figure=BarFigure(),
            cont_key=self._APP
        ))
        self.__controllers[self._APP][-1].layout.configuration.disable_subtitles()

        return charts
    
    def __get_assets_charts(self):
        charts: List[Callable] = list()

        charts.append(self.__create_sum_chart(
            title=TEXT['title_assets_value'],
            adapter=AssetsValues(),
            figure=TreemapFigure(),
            cont_key=self._ASSETS,
            dp_id=TEXT['dp_id_assets'],
            dp_title=TEXT['dp_title_assets'],
        ))
        self.__controllers[self._ASSETS][-1].layout.configuration.disable_subtitles()

        charts.append(self.__create_dataTable(
            title=TEXT['title_assets_novalue'],
            adapter=AssetsTokens(),
            cont_key=self._ASSETS
        ))

        return charts

    def __create_chart(self, title: str, adapter: IMetricAdapter, figure: Figure
    , cont_key: int, css_classes = []) -> Callable:
        """
        Creates the chart layout and its controller, and returns a callable
        to get the html representation.
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"
        layout: ChartPaneLayout = ChartPaneLayout(
            title=title,
            css_id=css_id,
            figure=figure,
            css_classes=css_classes,
        )
        layout.configuration.set_css_border(css_border=TEXT['css_pane_border'])

        controller: ChartController = ChartController(
            css_id=css_id,
            layout=layout,
            adapter=adapter)

        self.__controllers[cont_key].append(controller)
        return layout.get_layout

    def __create_sum_chart(self, 
        title: str,
        adapter: IMetricAdapter,
        figure: Figure,
        cont_key: int,
        dp_id: str,
        dp_title: str,
    ) -> Callable:
        """
        Creates the chart layout and its controller, and returns a callable
        to get the html representation.
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"
        layout: ChartPaneLayout = ChartPaneLayout(
            title=title,
            css_id=css_id,
            figure=figure
        )
        layout.configuration.set_css_border(css_border=TEXT['css_pane_border'])

        self.__data_points[dp_id] = DataPointLayout(
            css_id=dp_id,
            title=dp_title,
        )

        controller: ChartController = ChartSummaryController(
            css_id=css_id,
            layout=layout,
            adapter=adapter,
            datapoint_layout=self.__data_points[dp_id],
        )

        self.__controllers[cont_key].append(controller)
        return layout.get_layout

    def __create_dataTable(self, title: str, adapter: IMetricAdapter, cont_key: int) -> Callable:
        """Creates a datatable to put alongside charts

        Args:
            title (str): The title of the datatable
            adapter (IMetricAdapter): The adapter to get the data from
            cont_key (int): The key of the controller

        Returns:
            Callable: Layout html builder
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"

        layout: DataTableLayout = DataTableLayout(
            title=title,
            css_id=css_id
        )

        controller: DataTableController = DataTableController(
            table_id=layout.table_id,
            layout=layout,
            adapter=adapter
        )

        self.__controllers[cont_key].append(controller)
        return layout.get_layout



"""
   Descp: Tester for StTimeSerie.

   Created on: 14-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.apps.daostack.data_access.daos.metric.strategy.\
    st_time_serie import StTimeSerie

from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie


class StTimeSerieTest(unittest.TestCase):

    def __check_lists(self, df: pd.DataFrame, out: List[int]) -> None:
        # m_type it not used, in the future -1 will break it
        strategy: StTimeSerie = StTimeSerie(m_type=-1)
        result: StackedSerie = strategy.process_data(df=df)

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl.sub(month=3).change(day=1, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            {'id': '-1', 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-3)-01T00:00:00+00:00
            {'id': '3', 'trash': 'trash', 'createdAt': bl.add(month=1).change(day=25).unix()},#today_year-(today_month-2)-25T00:00:00+00:00
            {'id': '0', 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-2)-25T00:00:00+00:00
            {'id': '0', 'trash': 'trash', 'createdAt': bl.add(month=1).change(day=1, hour=23).unix()},#today_year-(today_month-1)-01T23:00:00+00:00
            {'id': '1', 'trash': 'trash', 'createdAt': bl.sub(month=1).change(minute=59, second=59).unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'id': '1', 'trash': 'trash', 'createdAt': bl.unix()},#today_year-(today_month-2)-01T23:59:59+00:00
            {'id': '2', 'trash': 'trash', 'createdAt': bl.add(month=2, second=1).change(day=21).unix()},#today_year-today_month-21T00:00:00+00:00
        ])

        out: List[int] = [1, 4, 1, 1]

        self.__check_lists(df=in_df, out=out)


if __name__ == "__main__":
    unittest.main()



"""
   Descp: Test for StProposalMajority class

   Created on: 20-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
import numpy as np
import unittest
from typing import List

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.apps.daostack.data_access.daos.metric.strategy.\
    st_proposal_majority import StProposalMajority
from dao_analyzer.apps.common.business.transfers.n_stacked_serie import NStackedSerie    

class StProposalMajorityTest(unittest.TestCase):
    
    def test_process_data(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        boost_date: int = bl.sub(month=3).change(day=28, hour=0, minute=0, second=0).unix()

        in_df: pd.DataFrame = pd.DataFrame([
            #executedAt = today_year-(today_month-3)-25T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.change(day=25).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '25', 'votesAgainst': '5', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-(today_month-3)-21T23:00:00+00:00; boostedAt = today_year-(today_month-2)-28T00:00:00+00:00
            {'executedAt': bl.change(day=21, hour=23).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '5', 'votesAgainst': '30', 'boostedAt': boost_date, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-(today_month-2)-01T12:00:00+00:00; boostedAt = today_year-(today_month-2)-28T00:00:00+00:00
            {'executedAt': bl.add(month=1).change(day=1, hour=12).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '5', 'votesAgainst': '1', 'boostedAt': boost_date, 'queuedVoteRequiredPercentage': '50'},

            #executedAt = today_year-(today_month-2)-15T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.change(day=15, hour=0).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '1', 'votesAgainst': '15', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},

            #executedAt = today_year-today_month-05T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.add(month=2).change(day=5).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '0', 'votesAgainst': '0', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
        ])
        out: List[List[int]] = [
            [50.0, np.nan, np.nan, np.nan], #abs_passes
            [np.nan, 10.0, np.nan, np.nan], #rel_passes
            [np.nan, 30.0, np.nan, 0.0], #rel_fails
            [60.0, np.nan, np.nan, np.nan] #abs_fails
        ]

        strategy: StProposalMajority = StProposalMajority()
        result: NStackedSerie = strategy.process_data(df=in_df)

        # numpy takes into consideration nan values
        for i, l in enumerate(out, 0):
            res_list: List[int] = result.get_i_stack(i_value=i, j_stack=0)
            np.testing.assert_array_equal(l, res_list)


if __name__ == "__main__":
    unittest.main()



"""
   Descp: Strategy pattern to create a metric of new additions.

   Created on: 19-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.serie import Serie
import dao_analyzer.apps.common.data_access.pandas_utils as pd_utl


class StNewAdditions(IMetricStrategy):
    VOTE = 0
    TRANSACTION = 1
    __TYPES: List[int] = [VOTE, TRANSACTION]

    __DF_VOTE_DATE = 'startDate'
    __DF_TRA_DATE = 'date'
    __DF_DATE = 'createdAt'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_COUNT]


    def __init__(self, typ: int) -> None:
        self.__typ: int = self.__get_type(typ)


    def __get_type(self, typ: int) -> int:
        """
        Checks if typ exists, if not return by default the first type
        """
        return typ if typ in self.__TYPES else self.__TYPES[0]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df

        if self.__typ is self.VOTE:
            dff = dff.rename(columns={self.__DF_VOTE_DATE: self.__DF_DATE})
        elif self.__typ is self.TRANSACTION:
            dff = dff.rename(columns={self.__DF_TRA_DATE: self.__DF_DATE})

        dff.loc[:, [self.__DF_DATE]] = dff[[self.__DF_DATE]]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        df = self.clean_df(df=df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        df = pd_utl.count_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)
        
        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)

        dff = pd_utl.get_df_from_lists([idx, 0], self.__DF_COLS)
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        # joinning all the data in a unique dataframe
        df = pd.concat([df, dff], ignore_index=True)
        df = df.drop_duplicates(subset=self.__DF_DATE, keep="first")
        df = df.sort_values(self.__DF_DATE)
        
        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric
